"""
VMC Distributed Firewall Rule state module

Add new distributed firewall rule, update existing distributed firewall rule and
     delete existing distributed firewall rule from an SDDC.

Example usage :

.. code-block:: yaml

    Distributed_Firewall_Rule_1:
      vmc_distributed_firewall_rules.present:
        - hostname: sample-nsx.vmwarevmc.com
        - refresh_key: 7jPSGSZpCa8e5Ouks4UY5cZyOtynAhF
        - authorization_host: console-stg.cloud.vmware.com
        - org_id: 10e1092f-51d0-473a-80f8-137652c39fd0
        - sddc_id: b43da080-2626-f64c-88e8-7f31d9d2c306
        - domain_id: default
        - security_policy_id: default-layer3-section
        - verify_ssl: False
        - cert: /path/to/client/certificate

.. warning::

    It is recommended to pass the VMC authentication details using Pillars rather than specifying as plain text in SLS
    files.


"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_state

log = logging.getLogger(__name__)


def present(
    name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_policy_id,
    verify_ssl=True,
    cert=None,
    source_groups=None,
    destination_groups=None,
    services=None,
    scope=None,
    action=None,
    sequence_number=None,
    display_name=None,
    disabled=None,
    logged=None,
    description=None,
    direction=None,
    notes=None,
    tag=None,
    tags=vmc_constants.VMC_NONE,
):
    """
    Ensure a given distributed firewall rule exists for given SDDC

    name
        Indicates the distributed firewall rule id, any unique string identifying the  distributed firewall rule.
        Also same as the display_name by default.

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the distributed firewall rules should be added

    domain_id
        The domain_id for which the Distributed firewall rule belongs to. Possible values: default, cgw

    security_policy_id
        The  security_policy_id for which the distributed firewall rules should belongs to

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    source_groups
        (Optional) List of Source group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    destination_groups
        (Optional) List of Destination group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    services
        (Optional) Names of services. In order to specify all services, use the constant "ANY".
        This is case insensitive. If "ANY" is used, it should be the ONLY element in the services array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    scope
        (Optional) The list of policy paths where the rule is applied LR/Edge/T0/T1/LRP etc.
        Note that a given rule can be applied on multiple LRs/LRPs.

    action
        (Optional) The action to be applied to all the services.
        Possible Values for are: ALLOW, DROP, REJECT

    display_name
        Identifier to use when displaying entity in logs or GUI
            Defaults to ID if not set

    sequence_number
        (Optional) Sequence number of the Rule.
        This field is used to resolve conflicts between multiple Rules under Security or Gateway Policy for a Domain.
        If no sequence number is specified by the user, a value of 0 is assigned by default.
        If there are multiple rules with the same sequence number then their order is not deterministic.
        If a specific order of rules is desired, then one has to specify unique sequence numbers.

    Disabled
        (Optional) Flag to disable the rule. Default is false.

    logged
        (Optional) Enable logging flag. Flag to enable packet logging. Default is disabled.

    description
        (Optional) Description of this resource

    direction
        (Optional) Define direction of traffic
        Possible Values for are: IN, OUT, IN_OUT
        Default: "IN_OUT"

    notes
        (Optional) Text for additional notes on changes.

    tag
        (Optional) Tag applied on the rule. User level field which will be printed in CLI and packet logs.

    tags
        (Optional) Opaque identifiers meaningful to the user.

        .. code-block::

            {
                "tag": <tag>,
                "scope": <scope>
            }

    Example values:

        .. code-block::

           {
              "description": "comm entry",
              "display_name": "",
              "sequence_number": 1,
              "source_groups": [
                "ANY"
              ],
              "logged": false,
              "destination_groups": [
                "ANY"
              ],
              "scope": [
                "ANY"
              ],
              "action": "DROP",
              "services": [
                "ANY"
              ],
              "direction": "IN_OUT",
              "tag": "",
              "notes": ""
            }

    """

    rule_id = name
    input_dict = {
        "source_groups": source_groups,
        "destination_groups": destination_groups,
        "services": services,
        "scope": scope,
        "action": action,
        "display_name": display_name,
        "sequence_number": sequence_number,
        "disabled": disabled,
        "logged": logged,
        "description": description,
        "direction": direction,
        "notes": notes,
        "tag": tag,
        "tags": tags,
    }

    input_dict = {k: v for k, v in input_dict.items() if v != vmc_constants.VMC_NONE}

    distributed_firewall_rule = __salt__["vmc_distributed_firewall_rules.get_by_id"](
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_policy_id=security_policy_id,
        rule_id=rule_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in distributed_firewall_rule:
        if "could not be found" in distributed_firewall_rule["error"]:
            distributed_firewall_rule = None
        else:
            return vmc_state._create_state_response(
                name=name, comment=distributed_firewall_rule["error"], result=False
            )

    if __opts__.get("test"):
        log.info("present is called with test option")
        return vmc_state._create_state_response(
            name=name,
            comment="Distributed firewall rule {} would have been {}".format(
                rule_id, "updated" if distributed_firewall_rule else "created"
            ),
        )

    if distributed_firewall_rule:
        updatable_keys = input_dict.keys()
        is_update_required = vmc_state._check_for_updates(
            distributed_firewall_rule, input_dict, updatable_keys, ["tags"]
        )

        if is_update_required:
            updated_distributed_firewall_rule = __salt__["vmc_distributed_firewall_rules.update"](
                hostname=hostname,
                refresh_key=refresh_key,
                authorization_host=authorization_host,
                org_id=org_id,
                sddc_id=sddc_id,
                domain_id=domain_id,
                security_policy_id=security_policy_id,
                rule_id=rule_id,
                verify_ssl=verify_ssl,
                cert=cert,
                source_groups=source_groups,
                destination_groups=destination_groups,
                services=services,
                scope=scope,
                action=action,
                display_name=display_name,
                sequence_number=sequence_number,
                disabled=disabled,
                logged=logged,
                description=description,
                direction=direction,
                notes=notes,
                tag=tag,
                tags=tags,
            )

            if "error" in updated_distributed_firewall_rule:
                return vmc_state._create_state_response(
                    name=name, comment=updated_distributed_firewall_rule["error"], result=False
                )

            updated_distributed_firewall_rule = __salt__[
                "vmc_distributed_firewall_rules.get_by_id"
            ](
                hostname=hostname,
                refresh_key=refresh_key,
                authorization_host=authorization_host,
                org_id=org_id,
                sddc_id=sddc_id,
                domain_id=domain_id,
                security_policy_id=security_policy_id,
                rule_id=rule_id,
                verify_ssl=verify_ssl,
                cert=cert,
            )

            if "error" in updated_distributed_firewall_rule:
                return vmc_state._create_state_response(
                    name=name,
                    comment=updated_distributed_firewall_rule["error"],
                    result=False,
                )

            return vmc_state._create_state_response(
                name=name,
                comment="Updated distributed firewall rule {}".format(rule_id),
                old_state=distributed_firewall_rule,
                new_state=updated_distributed_firewall_rule,
                result=True,
            )
        else:
            log.info("All fields are same as existing distributed firewall rule %s", rule_id)
            return vmc_state._create_state_response(
                name=name,
                comment="Distributed firewall rule exists already, no action to perform",
                result=True,
            )
    else:
        log.info("No distributed firewall rule found with ID %s", rule_id)
        created_distributed_firewall_rule = __salt__["vmc_distributed_firewall_rules.create"](
            hostname=hostname,
            refresh_key=refresh_key,
            authorization_host=authorization_host,
            org_id=org_id,
            sddc_id=sddc_id,
            domain_id=domain_id,
            security_policy_id=security_policy_id,
            rule_id=rule_id,
            verify_ssl=verify_ssl,
            cert=cert,
            source_groups=source_groups,
            destination_groups=destination_groups,
            services=services,
            scope=scope,
            action=action,
            sequence_number=sequence_number,
            disabled=disabled,
            logged=logged,
            description=description,
            direction=direction,
            notes=notes,
            tag=tag,
            tags=tags,
        )

        if "error" in created_distributed_firewall_rule:
            return vmc_state._create_state_response(
                name=name, comment=created_distributed_firewall_rule["error"], result=False
            )

        return vmc_state._create_state_response(
            name=name,
            comment="Created distributed firewall rule {}".format(rule_id),
            new_state=created_distributed_firewall_rule,
            result=True,
        )


def absent(
    name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_policy_id,
    verify_ssl=True,
    cert=None,
):
    """
    Ensure a given distributed firewall rule does not exist on given SDDC

    name
        Indicates the distributed firewall rule id, any unique string identifying the distributed firewall rule.

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC from which the distributed firewall rule should be deleted

    domain_id
        The domain_id for which the Distributed firewall rule belongs to. Possible values: default, cgw

    security_policy_id
        The  security_policy_id for which the distributed firewall rules belongs to

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    rule_id = name
    log.info("Checking if distributed firewall rule with ID %s is present", rule_id)
    distributed_firewall_rule = __salt__["vmc_distributed_firewall_rules.get_by_id"](
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        security_policy_id=security_policy_id,
        domain_id=domain_id,
        rule_id=rule_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    if "error" in distributed_firewall_rule:
        if "could not be found" in distributed_firewall_rule["error"]:
            distributed_firewall_rule = None
        else:
            return vmc_state._create_state_response(
                name=name, comment=distributed_firewall_rule["error"], result=False
            )

    if __opts__.get("test"):
        log.info("absent is called with test option")
        if distributed_firewall_rule:
            return vmc_state._create_state_response(
                name=name,
                comment="State absent will delete distributed firewall rule with ID {}".format(
                    rule_id
                ),
            )
        else:
            return vmc_state._create_state_response(
                name=name,
                comment="State absent will do nothing as no distributed firewall rule found with ID {}".format(
                    rule_id
                ),
            )

    if distributed_firewall_rule:
        log.info("Distributed firewall rule found with ID %s", rule_id)
        deleted_distributed_firewall_rule = __salt__["vmc_distributed_firewall_rules.delete"](
            hostname=hostname,
            refresh_key=refresh_key,
            authorization_host=authorization_host,
            org_id=org_id,
            sddc_id=sddc_id,
            domain_id=domain_id,
            security_policy_id=security_policy_id,
            rule_id=rule_id,
            verify_ssl=verify_ssl,
            cert=cert,
        )

        if "error" in deleted_distributed_firewall_rule:
            return vmc_state._create_state_response(
                name=name, comment=deleted_distributed_firewall_rule["error"], result=False
            )

        return vmc_state._create_state_response(
            name=name,
            comment="Deleted distributed firewall rule {}".format(rule_id),
            old_state=distributed_firewall_rule,
            result=True,
        )
    else:
        log.info("No distributed firewall rule found with ID %s", rule_id)
        return vmc_state._create_state_response(
            name=name,
            comment="No distributed firewall rule found with ID {}".format(rule_id),
            result=True,
        )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging

import salt.exceptions
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.datacenter as utils_datacenter
import saltext.vmware.utils.esxi as utils_esxi
import saltext.vmware.utils.vsphere as utils_vmware

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim, vmodl

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

__virtualname__ = "vmware_dvswitch"
__proxyenabled__ = ["vmware_dvswitch"]
__func_alias__ = {"list_": "list"}


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def _get_switch_config_spec(service_instance, datacenter_name, switch_name):
    dc_ref = switch_ref = config_spec = None
    dc_ref = utils_datacenter.get_datacenter(service_instance, datacenter_name)
    switch_refs = utils_vmware.get_dvss(dc_ref=dc_ref, dvs_names=[switch_name])
    if switch_refs:
        switch_ref = switch_refs[0]
        dvs_props = utils_common.get_properties_of_managed_object(
            switch_ref, ["config", "capability"]
        )
        config_spec = vim.dvs.VmwareDistributedVirtualSwitch.ConfigSpec()
        # Copy all of the properties in the config of the of the DVS to a
        # DvsConfigSpec
        skipped_properties = ["host"]
        for prop in config_spec.__dict__.keys():
            if prop in skipped_properties:
                continue
            if hasattr(dvs_props["config"], prop):
                setattr(config_spec, prop, getattr(dvs_props["config"], prop))
    return dc_ref, switch_ref, config_spec


def configure(
    datacenter_name,
    switch_name,
    uplink_count=None,
    uplink_prefix="Uplink ",
    switch_version=None,
    switch_description=None,
    mtu=None,
    discovery_protocol=None,
    discovery_operation=None,
    multicast_filtering_mode=None,
    contact_name=None,
    contact_description=None,
    network_forged_transmits=None,
    network_mac_changes=None,
    network_promiscuous=None,
    health_check_teaming_failover=None,
    health_teaming_failover_interval=None,
    health_vlan_mtu=None,
    health_vlan_mtu_interval=None,
    service_instance=None,
    profile=None,
):
    """
    Creates a new distributed vSwitch or updates an existing vSwitch.

    switch_name
        Name of the distributed vSwitch to create or update.

    uplink_count
        Count of uplink per ESXi per host. Optional.

    uplink_prefix
        The prefix to be used for uplinks. Optional. Default: "Uplink ".

    switch_version
        The version of the distributed vSwitch to create or update. Optional.

    switch_description
        Description of the distributed vSwitch. Optional. Default: None.

    mtu
        Maximum transmission unit for the switch. Optional.

    discovery_protocol
        Link discovery protocol between Cisco and Link Layer discovery. Optional. Valid values: "cdp", "lldp", "disabled".

    discovery_operation
        Discovery operation for the switch. Optional. Valid values: "both", "advertise", "listen".

    multicast_filtering_mode
        Multicast filtering mode for the switch. Optional. Valid values: "basic", "snooping".

    contact_name
        Administrator contact name. Optional. Default: "".

    contact_description
        Administrator contact information. Optional. Default: "".

    network_forged_transmits
        Allow forged transmits. Type: Boolean. Optional. Valid values: "True", "False".

    network_mac_changes
        Allow mac changes. Type: Boolean. Optional. Valid values: "True", "False".

    network_promiscuous
        Allow promiscuous mode. Type: Boolean. Optional. Valid values: "True", "False".

    health_check_teaming_failover
        Enable teaming and failover health check. Type: Boolean. Optional. Valid values: "True", "False".

    health_teaming_failover_interval
        Teaming and failover health check interval in minutes. Optional.

    health_vlan_mtu
        Enable VLAN and MTU health check. Type: Boolean. Optional. Valid values: "True", "False".

    health_vlan_mtu_interval
        VLAN and MTU health check interval in minutes. Optional.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: bash

        salt '*' vmware_dvswitch.configure dvs1
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    try:
        health_spec = product_spec = spec = None
        dc_ref, switch_ref, config_spec = _get_switch_config_spec(
            service_instance=service_instance,
            datacenter_name=datacenter_name,
            switch_name=switch_name,
        )
        if not switch_ref:
            spec = vim.DistributedVirtualSwitch.CreateSpec()
            spec.configSpec = vim.dvs.VmwareDistributedVirtualSwitch.ConfigSpec()
            config_spec = spec.configSpec
            config_spec.name = switch_name

        if mtu:
            config_spec.maxMtu = mtu

        if uplink_count:
            config_spec.uplinkPortPolicy = vim.DistributedVirtualSwitch.NameArrayUplinkPortPolicy()
            for i in range(uplink_count):
                config_spec.uplinkPortPolicy.uplinkPortName.append(
                    "{}{}".format(uplink_prefix, i + 1)
                )

        if switch_version:
            product_spec = vim.dvs.ProductSpec()
            product_spec.version = switch_version
            if spec:
                spec.productInfo = product_spec

        if switch_description:
            config_spec.description = switch_description

        if contact_name or contact_description:
            contact_info_spec = vim.DistributedVirtualSwitch.ContactInfo()
            contact_info_spec.contact = contact_description
            contact_info_spec.name = contact_name
            if switch_ref:
                contact_info_spec.contact = contact_description or switch_ref.config.contact.contact
                contact_info_spec.name = contact_name or switch_ref.config.contact.name
            config_spec.contact = contact_info_spec

        if discovery_operation or discovery_protocol:
            ldp_config_spec = vim.host.LinkDiscoveryProtocolConfig()
            ldp_config_spec.operation = discovery_operation
            ldp_config_spec.protocol = discovery_protocol
            if switch_ref:
                ldp_config_spec.protocol = (
                    discovery_protocol or switch_ref.config.linkDiscoveryProtocolConfig.protocol
                )
                ldp_config_spec.operation = (
                    discovery_operation or switch_ref.config.linkDiscoveryProtocolConfig.operation
                )
            if discovery_protocol == "disabled":
                ldp_config_spec.protocol = "cdp"
                ldp_config_spec.operation = "none"
            config_spec.linkDiscoveryProtocolConfig = ldp_config_spec

        if multicast_filtering_mode:
            if multicast_filtering_mode == "basic":
                config_spec.multicastFilteringMode = "legacyFiltering"
            else:
                config_spec.multicastFilteringMode = multicast_filtering_mode

        if not switch_ref:
            utils_vmware.create_dvs(dc_ref=dc_ref, dvs_name=switch_name, dvs_create_spec=spec)

        if (
            network_promiscuous is not None
            or network_mac_changes is not None
            or network_forged_transmits is not None
        ):
            policy = vim.dvs.VmwareDistributedVirtualSwitch.SecurityPolicy()
            if not switch_ref:
                dc_ref, switch_ref, config_spec = _get_switch_config_spec(
                    service_instance=service_instance,
                    datacenter_name=datacenter_name,
                    switch_name=switch_name,
                )
            if network_promiscuous is None:
                network_promiscuous = (
                    switch_ref.config.defaultPortConfig.securityPolicy.allowPromiscuous.value
                )
            if network_mac_changes is None:
                network_mac_changes = (
                    switch_ref.config.defaultPortConfig.securityPolicy.macChanges.value
                )
            if network_forged_transmits is None:
                network_forged_transmits = (
                    switch_ref.config.defaultPortConfig.securityPolicy.forgedTransmits.value
                )
            if network_promiscuous is not None:
                policy.allowPromiscuous = vim.BoolPolicy(value=network_promiscuous)
            if network_mac_changes is not None:
                policy.macChanges = vim.BoolPolicy(value=network_mac_changes)
            if network_forged_transmits is not None:
                policy.forgedTransmits = vim.BoolPolicy(value=network_forged_transmits)
            config_spec.defaultPortConfig.securityPolicy = policy

        if (
            health_check_teaming_failover is not None
            or health_teaming_failover_interval is not None
            or health_vlan_mtu is not None
            or health_vlan_mtu_interval is not None
        ):
            if not switch_ref:
                dc_ref, switch_ref, config_spec = _get_switch_config_spec(
                    service_instance=service_instance,
                    datacenter_name=datacenter_name,
                    switch_name=switch_name,
                )
            health_spec = vim.DistributedVirtualSwitch.HealthCheckConfig.Array()
            for config in switch_ref.config.healthCheckConfig:
                if isinstance(
                    config, vim.dvs.VmwareDistributedVirtualSwitch.VlanMtuHealthCheckConfig
                ):
                    if health_vlan_mtu is not None:
                        config.enable = health_vlan_mtu
                    if health_vlan_mtu_interval is not None:
                        config.interval = health_vlan_mtu_interval
                    health_spec.append(config)
                if isinstance(
                    config, vim.dvs.VmwareDistributedVirtualSwitch.TeamingHealthCheckConfig
                ):
                    if health_check_teaming_failover is not None:
                        config.enable = health_check_teaming_failover
                    if health_teaming_failover_interval is not None:
                        config.interval = health_teaming_failover_interval
                    health_spec.append(config)

        if switch_ref:
            utils_vmware.update_dvs(dvs_ref=switch_ref, dvs_config_spec=config_spec)
            if product_spec:
                utils_vmware.update_dvs_version(dvs_ref=switch_ref, dvs_product_spec=product_spec)
            if health_spec:
                utils_vmware.update_dvs_health(dvs_ref=switch_ref, dvs_health_spec=health_spec)

        return True
    except (
        vim.fault.DvsFault,
        vmodl.fault.NotSupported,
        salt.exceptions.VMwareApiError,
        vmodl.RuntimeFault,
        vmodl.MethodFault,
    ) as exc:
        raise salt.exceptions.SaltException(str(exc))


def remove_hosts(
    switch_name,
    host_name=None,
    datacenter_name=None,
    cluster_name=None,
    service_instance=None,
    profile=None,
):
    """
    Remove ESXi host(s) from a distributed vSwitch.

    switch_name
        Name of the distributed vSwitch.

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: bash

        salt '*' vmware_dvswitch.remove_hosts switch_name=dvs1 hostname=host1

    """
    log.debug("Running vmware_dvswitch.remove_hosts")
    ret = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )
    try:
        for h in hosts:
            ret[h.name] = False
            if hasattr(h.configManager.networkSystem.networkInfo, "vswitch"):
                h.configManager.networkSystem.RemoveVirtualSwitch(vswitchName=switch_name)
                ret[h.name] = True
        return ret
    except (
        vim.fault.InvalidState,
        vim.fault.NotFound,
        vim.fault.HostConfigFault,
        vim.fault.ResourceInUse,
        vmodl.fault.InvalidArgument,
        salt.exceptions.VMwareApiError,
    ) as exc:
        raise salt.exceptions.SaltException(str(exc))


def add_hosts(
    switch_name,
    host_name=None,
    datacenter_name=None,
    cluster_name=None,
    nics=None,
    service_instance=None,
    profile=None,
):
    """
    Add ESXi host(s) to a distributed vSwitch.

    switch_name
        Name of the distributed vSwitch.

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    nics
        List of vmnics to attach to vSwitch. (optional). Default "None".

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: bash

        salt '*' vmware_dvswitch.add_hosts switch_name=dvs1 host_name=host1 num_ports=256 mtu=1800

    """
    log.debug("Running vmware_dvswitch.add_hosts")
    ret = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    _, switch_ref, _ = _get_switch_config_spec(
        service_instance=service_instance,
        datacenter_name=datacenter_name,
        switch_name=switch_name,
    )
    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )
    try:
        for h in hosts:
            ret[h.name] = False
            network_manager = h.configManager.networkSystem
            if not network_manager:
                continue
            if isinstance(nics, str):
                nics = [nics]
            vss_spec = vim.host.VirtualSwitch.Specification()
            vss_spec.mtu = switch_ref.config.maxMtu or 1500
            vss_spec.numPorts = switch_ref.config.numPorts or 128
            if nics:
                vss_spec.bridge = vim.host.VirtualSwitch.BondBridge(nicDevice=nics)
            network_manager.AddVirtualSwitch(vswitchName=switch_name, spec=vss_spec)
            ret[h.name] = True
        return ret
    except (
        vim.fault.InvalidState,
        vim.fault.NotFound,
        vim.fault.HostConfigFault,
        vim.fault.ResourceInUse,
        vmodl.fault.InvalidArgument,
        salt.exceptions.VMwareApiError,
    ) as exc:
        raise salt.exceptions.SaltException(str(exc))


def update_hosts(
    switch_name,
    host_name=None,
    datacenter_name=None,
    cluster_name=None,
    nics=None,
    num_ports=None,
    mtu=None,
    service_instance=None,
    profile=None,
):
    """
    Update ESXi host(s) on a distributed vSwitch.

    switch_name
        Name of the distributed vSwitch.

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this ESXi hostname (optional)

    nics
        List of vmnics to attach to vSwitch. (optional). Default "None".

    num_ports
        Number of ports to configure on the vSwitch. (optional). Default 128.

    mtu
        MTU to configure on the vSwitch. (optional). Default 1600.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    .. code-block:: bash

        salt '*' vmware_dvswitch.update_hosts switch_name=dvs1 host_name=host1 num_ports=256 mtu=1800

    """
    log.debug("Running vmware_dvswitch.update_hosts")
    ret = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    _, switch_ref, _ = _get_switch_config_spec(
        service_instance=service_instance,
        datacenter_name=datacenter_name,
        switch_name=switch_name,
    )
    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )
    try:
        for h in hosts:
            ret[h.name] = False
            network_manager = h.configManager.networkSystem
            if not network_manager:
                continue
            if isinstance(nics, str):
                nics = [nics]
            vss_spec = vim.host.VirtualSwitch.Specification()
            if hasattr(network_manager.networkInfo, "vswitch"):
                for switch in network_manager.networkInfo.vswitch:
                    if switch.name != switch_name:
                        continue
                    vss_spec.mtu = mtu or switch.mtu or switch_ref.config.maxMtu
                    vss_spec.numPorts = (
                        num_ports or switch.spec.numPorts or switch_ref.config.numPorts
                    )
                    if switch.pnic or nics:
                        vss_spec.bridge = vim.host.VirtualSwitch.BondBridge(
                            nicDevice=nics or list(map(lambda x: x.split("-", 3)[-1], switch.pnic))
                        )
                    network_manager.UpdateVirtualSwitch(vswitchName=switch_name, spec=vss_spec)
                    ret[h.name] = True
        return ret
    except (
        vim.fault.InvalidState,
        vim.fault.NotFound,
        vim.fault.HostConfigFault,
        vim.fault.ResourceInUse,
        vmodl.fault.InvalidArgument,
        salt.exceptions.VMwareApiError,
    ) as exc:
        raise salt.exceptions.SaltException(str(exc))


def list_(
    service_instance=None,
    profile=None,
):
    """
    .. versionadded:: <<VERSION>>

    Returns a list of distributed virtual switches for the specified host.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_dvswitch.list
    """
    log.debug("Running vmware_dvswitch.list")
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    return utils_vmware.list_dvs(service_instance)



"""
Salt execution module for VMC vm disks
Provides methods to Create, Read, Update and Delete virtual disks from the VM.
"""
import logging
import os

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request
from saltext.vmware.utils import vmc_templates
from saltext.vmware.utils import vmc_vcenter_request

log = logging.getLogger(__name__)

__virtualname__ = "vmc_vm_disks"
__func_alias__ = {"list_": "list"}


def __virtual__():
    return __virtualname__


def list_(hostname, username, password, vm_id, verify_ssl=True, cert=None):
    """
    Retrieves the available virtual disks for given VM.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vm_disks.list hostname=sample-vcenter.vmwarevmc.com ...

    hostname
        Hostname of the vCenter console

    username
        username required to login to vCenter console

    password
        password required to login to vCenter console

    vm_id
        Virtual machine identifier for which the available hard disks should be retrieved.

    verify_ssl
        (optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving the available virtual disks for VM %s", vm_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = "{base_url}api/vcenter/vm/{vm_id}/hardware/disk"
    api_url = api_url.format(base_url=api_url_base, vm_id=vm_id)

    headers = vmc_vcenter_request.get_headers(hostname, username, password)
    return vmc_vcenter_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        headers=headers,
        description="vmc_vm_disks.list",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get(hostname, username, password, vm_id, disk_id, verify_ssl=True, cert=None):
    """
    Retrieves details of given virtual disk for given VM.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vm_disks.get hostname=sample-vcenter.vmwarevmc.com ...

    hostname
        Hostname of the vCenter console

    username
        username required to login to vCenter console

    password
        password required to login to vCenter console

    vm_id
        Virtual machine identifier for which the disk details should be retrieved.

    disk_id
        Virtual disk identifier for which the details should be retrieved from given VM

    verify_ssl
        (optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving details of virtual disk %s for VM %s", disk_id, vm_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = "{base_url}api/vcenter/vm/{vm_id}/hardware/disk/{disk_id}"
    api_url = api_url.format(base_url=api_url_base, vm_id=vm_id, disk_id=disk_id)

    headers = vmc_vcenter_request.get_headers(hostname, username, password)
    return vmc_vcenter_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        headers=headers,
        description="vmc_vm_disks.get",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def delete(hostname, username, password, vm_id, disk_id, verify_ssl=True, cert=None):
    """
    Deletes given virtual disk from given VM.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vm_disks.delete hostname=sample-vcenter.vmwarevmc.com ...

    hostname
        Hostname of the vCenter console

    username
        username required to login to vCenter console

    password
        password required to login to vCenter console

    vm_id
        Virtual machine identifier for which the disk should be removed.

    disk_id
        Virtual disk identifier which has to be removed from given VM

    verify_ssl
        (optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Deleting virtual disk %s from VM %s", disk_id, vm_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = "{base_url}api/vcenter/vm/{vm_id}/hardware/disk/{disk_id}"
    api_url = api_url.format(base_url=api_url_base, vm_id=vm_id, disk_id=disk_id)

    headers = vmc_vcenter_request.get_headers(hostname, username, password)
    return vmc_vcenter_request.call_api(
        method=vmc_constants.DELETE_REQUEST_METHOD,
        url=api_url,
        headers=headers,
        description="vmc_vm_disks.delete",
        responsebody_applicable=False,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def update(
    hostname,
    username,
    password,
    vm_id,
    disk_id,
    verify_ssl=True,
    cert=None,
    backing_type=None,
    vmdk_file=None,
):
    """
    Updates the configuration of a given virtual disk of given VM. An update operation can be used to detach
    the existing VMDK file and attach another VMDK file to the virtual machine.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vm_disks.update hostname=sample-vcenter.vmwarevmc.com ...

    hostname
        Hostname of the vCenter console

    username
        username required to login to vCenter console

    password
        password required to login to vCenter console

    vm_id
        Virtual machine identifier for which the disk should be added.

    disk_id
        Virtual disk identifier which has to be updated for given VM

    verify_ssl
        (optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    backing_type
        (mandatory) Backing type for the virtual disk. Possible values are: VMDK_FILE

    vmdk_file
        (mandatory) Path of the VMDK file backing the virtual disk.

    """

    log.info("Updating configuration for virtual disk %s of VM %s", disk_id, vm_id)

    validation_result = []
    if backing_type is None:
        validation_result.append("backing_type")
    if vmdk_file is None:
        validation_result.append("vmdk_file")

    if validation_result:
        error_msg = "Mandatory params {} are missing from user input".format(validation_result)
        log.error(error_msg)
        return {vmc_constants.ERROR: error_msg}

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = "{base_url}api/vcenter/vm/{vm_id}/hardware/disk/{disk_id}"
    api_url = api_url.format(base_url=api_url_base, vm_id=vm_id, disk_id=disk_id)

    headers = vmc_vcenter_request.get_headers(hostname, username, password)

    payload = {"backing": {"type": backing_type, "vmdk_file": vmdk_file}}

    return vmc_vcenter_request.call_api(
        method=vmc_constants.PATCH_REQUEST_METHOD,
        url=api_url,
        headers=headers,
        description="vmc_vm_disks.update",
        responsebody_applicable=False,
        verify_ssl=verify_ssl,
        cert=cert,
        data=payload,
    )


def create(
    hostname,
    username,
    password,
    vm_id,
    bus_adapter_type,
    verify_ssl=True,
    cert=None,
    vmdk=None,
    capacity=None,
    storage_policy_id=None,
    scsi=None,
    sata=None,
    ide=None,
):
    """
    Creates virtual disk for given VM.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vm_disks.create hostname=sample-vcenter.vmwarevmc.com ...

    hostname
       Hostname of the vCenter console

    username
        username required to login to vCenter console

    password
        password required to login to vCenter console

    vm_id
        Virtual machine identifier for which the disk should be added.

    bus_adapter_type
        Type of host bus adapter to which the disk should be attached. Possible values are: IDE, SATA, SCSI

    verify_ssl
        (optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    vmdk
        (optional) If empty, ``None``, or a string create a new VMDK.
        If empty or ``None``, creates new VMDK with a name (derived from the name of the virtual machine) chosen by the server.
        If vmdk ends with ``.vmdk``, it must refer to an existing VMDK.
        Otherwise, ``vmdk`` will be the stem of the new VMDK.
        Example ``vmdk="test-vm-1"`` will result in ``[WorkloadDatastore] 8dff8760-b6da-24d2-475b-0200a629f7fc/test-vm-1.vmdk``.

    capacity
        (Optional) The capacity for the disk, in bytes. If unset, defaults to a guest-specific capacity.
        Prohibited if the ``vmdk`` ends with ``.vmdk``

    storage_policy_id
        (Optional) The identifier for the storage policy to be used when creating the vmdk file. Prohibited if
        the ``vmdk`` ends with ``.vmdk``

    scsi
        (optional) Address for attaching the device to a virtual SCSI adapter. If unset, the server will choose
        an available address; if none is available, the request will fail.

        It is a json object which can contain 'bus' and 'unit' keys.

        'bus': (Integer As Int64) (mandatory)
            Bus number of the adapter to which the device should be attached.

        'unit': (Integer As Int64) (optional)
            Unit number of the device. If unset, the server will choose an available unit number on the
            specified adapter. If there are no available connections on the adapter, the request will be rejected.

        For ex:

            .. code::

                "scsi": {
                    "bus": 0,
                    "unit": 1
                }

    sata
        (optional) Address for attaching the device to a virtual SATA adapter. If unset, the server will choose
        an available address; if none is available, the request will fail.

        It is a json object which can contain 'bus' and 'unit' keys.

        'bus': (Integer As Int64) (mandatory)
            Bus number of the adapter to which the device should be attached.

        'unit': (Integer As Int64) (optional)
            Unit number of the device. If unset, the server will choose an available unit number on the
            specified adapter. If there are no available connections on the adapter, the request will be rejected.

        For ex:

            .. code::

                "sata": {
                    "bus": 0,
                    "unit": 1
                }

    ide
        (optional) Address for attaching the device to a virtual IDE adapter. If unset, the server will choose
        an available address; if none is available, the request will fail.

        It is a json object which can contain 'master' and 'primary' keys.

        'master': (Boolean) (Optional)
            Flag specifying whether the device should be the master or slave device on the IDE adapter.
            If unset, the server will choose an available connection type. If no IDE connections are available,
            the request will be rejected.

        'primary': (Boolean) (Optional)
            Flag specifying whether the device should be attached to the primary or secondary IDE adapter
            of the virtual machine. If unset, the server will choose a adapter with an available connection.
            If no IDE connections are available, the request will be rejected.

        For ex:

            .. code::

                "ide": {
                    "master": false,
                    "primary": false
                }

    """

    log.info("Creating virtual disk on virtual %s adapter for VM %s", bus_adapter_type, vm_id)

    backing = None
    new_vmdk = None
    if vmdk and vmdk.endswith(".vmdk"):
        if capacity is not None or storage_policy_id is not None:
            error_msg = "If vmdk={!r} ends in '.vmdk' then capacity and storage_policy_id cannot be specified".format(
                vmdk
            )
            log.error(error_msg)
            return {vmc_constants.ERROR: error_msg}
        backing = {"type": "VMDK_FILE", "vmdk_file": vmdk}
    else:
        new_vmdk = {}
        if vmdk:
            new_vmdk["name"] = vmdk
        if capacity:
            new_vmdk["capacity"] = int(capacity)
        if storage_policy_id:
            new_vmdk["storage_policy"] = {"policy": storage_policy_id}

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = "{base_url}api/vcenter/vm/{vm_id}/hardware/disk"
    api_url = api_url.format(base_url=api_url_base, vm_id=vm_id)

    headers = vmc_vcenter_request.get_headers(hostname, username, password)

    allowed_dict = {
        "backing": backing,
        "ide": ide,
        "new_vmdk": new_vmdk,
        "sata": sata,
        "scsi": scsi,
    }

    req_data = vmc_request._filter_kwargs(allowed_kwargs=allowed_dict.keys(), **allowed_dict)

    payload = vmc_request.create_payload_for_request(vmc_templates.create_vm_disks, req_data)
    payload["type"] = bus_adapter_type

    return vmc_vcenter_request.call_api(
        method=vmc_constants.POST_REQUEST_METHOD,
        url=api_url,
        headers=headers,
        description="vmc_vm_disks.create",
        verify_ssl=verify_ssl,
        cert=cert,
        data=payload,
    )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging


log = logging.getLogger(__name__)

try:
    import pyVmomi

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "vmware_license_mgr"
__proxyenabled__ = ["vmware_license_mgr"]


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def absent(license_key, **kwargs):
    """
    Remove a license specified by license_key from a Datacenter, Cluster, ESXI Server or vCenter
    If no datacenter, cluster or ESXI Server is specified, it is assumed the operation is to be applied to a vCenter

    license_key
        License Key which specifies license to remove from license manager

    datacenter_name
        Datacenter name to use for the operation [default None]

    cluster_name
        Name of the cluster from which to remove license [default None]

    esxi_hostname
        Hostname of the ESXI Server from which to remove the license [default None]

    CLI Example:

    .. code-block:: yaml

        Remove license from License Manager:
          vmware_license_mgr.absent:
            - license_key: license_key
            - datacenter_name: dc1

    """
    ret = __salt__["vmware_license_mgr.remove"](license_key, **kwargs)
    return ret


def present(license_key, **kwargs):
    """
    Add a license specified by license key to a Datacenter, Cluster, ESXI Server or vCenter
    If no datacenter, cluster or ESXI Server is specified, it is assumed the operation is to be applied to a vCenter

    license_key
        License Key which specifies license to add to license manager

    datacenter_name
        Datacenter name to use for the operation [default None]

    cluster_name
        Name of the cluster to add license [default None]

    esxi_hostname
        Hostname of the ESXI Server to add license [default None]

    CLI Example:

    .. code-block:: yaml

        Add license to License Manager:
          vmware_license_mgr.present:
            - license_key: license_key
            - datacenter_name: dc1
    """
    ret = __salt__["vmware_license_mgr.add"](license_key, **kwargs)
    return ret



"""
Salt execution module for VMC DNS Forwarder zones and services
Provides methods to show DNS zones and services.
"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request

log = logging.getLogger(__name__)

__virtualname__ = "vmc_dns_forwarder"


def __virtual__():
    return __virtualname__


def get_dns_zones(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    verify_ssl=True,
    cert=None,
    sort_by=None,
    sort_ascending=None,
    page_size=None,
    cursor=None,
):
    """
    Retrieves DNS zones for Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_dns_forwarder.get_dns_zones hostname=nsxt-manager.local...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the DNS zones should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order. Enabled by default.

    page_size
        (Optional) Maximum number of results to return in this page. Default page size is 1000.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    """

    log.info("Retrieving DNS zones for SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/dns-forwarder-zones"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id)

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["cursor", "page_size", "sort_ascending", "sort_by"],
        cursor=cursor,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_dns_forwarder.get_dns_zones",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_dns_services(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    verify_ssl=True,
    cert=None,
    page_size=None,
    cursor=None,
):
    """
    Retrieves DNS services for the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_dns_forwarder.get_dns_services hostname=nsxt-manager.local ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the DNS services should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    page_size
        (Optional) Maximum number of results to return in this page. Default page size is 1000.

    cursor
        (Optional) Integer cursor to be used for getting next page of records (supplied by current result page)

    """

    log.info("Retrieving DNS services for SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/search?query=resource_type:PolicyDnsForwarder"
    )

    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id)

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["cursor", "page_size"], cursor=cursor, page_size=page_size
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_dns_forwarder.get_dns_services",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )



"""
Salt execution Module to manage VMware NSX-T transport-zones
"""
import logging

from saltext.vmware.utils import common
from saltext.vmware.utils import nsxt_request

log = logging.getLogger(__name__)
__virtual_name__ = "nsxt_transport_zone"


def __virtual__():
    return __virtual_name__


TRANSPORT_ZONE_BASE_URL = "https://{0}/api/v1/transport-zones"

create_params_for_transport_zones = [
    "host_switch_name",
    "transport_type",
    "description",
    "host_switch_id",
    "host_switch_mode",
    "uplink_teaming_policy_names",
    "tags",
    "is_default",
    "display_name",
]


def get(
    hostname,
    username,
    password,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    cursor=None,
    included_fields=None,
    page_size=None,
    sort_ascending=None,
    sort_by=None,
):
    """
    Retrieves transport zones from Given NSX-T Manager

    CLI Example:
    .. code-block:: bash

        salt vm_minion nsxt_transport_zones.get_transport_zone hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    include_system_owned
        (Optional) Filter to indicate whether to include system owned Transport Zones.

    included_fields
        (Optional) Comma separated list of fields that should be included in query result

    is_default
        (Optional) Filter to choose if default transport zones will be returned.
        If set to true, only the default transport zones will be returned. If set to false, all transport zones except
        the default ones will be returned. If unset, all transport zones will be returned.

    page_size
        (Optional) Maximum number of results to return in this page (server may return fewer)

    sort_ascending
        (Optional) Value to sort result in ascending order

    sort_by
        (Optional) Field by which records are sorted

    """
    log.info("Fetching transport zones")
    url = TRANSPORT_ZONE_BASE_URL.format(hostname)
    params = common._filter_kwargs(
        allowed_kwargs=["cursor", "included_fields", "page_size", "sort_ascending", "sort_by"],
        default_dict=None,
        cursor=cursor,
        included_fields=included_fields,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )
    return nsxt_request.call_api(
        method="get",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_by_display_name(
    hostname,
    username,
    password,
    display_name,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
):
    """
    Lists the Transport Zones based on display name in the NSX-T Manager

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_node.get_transport_node_by_display_name hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    included_fields
        (Optional) Comma separated list of fields that should be included in query result

    page_size
        (Optional) Maximum number of results to return in this page

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order
    """
    log.info("Finding transport zones with display name: %s", display_name)
    transport_zones = common._read_paginated(
        func=get,
        display_name=display_name,
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )
    if "error" in transport_zones:
        return transport_zones
    return {"results": transport_zones}


def create(
    hostname,
    username,
    password,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    host_switch_name=None,
    transport_type=None,
    description=None,
    display_name=None,
    host_switch_id=None,
    host_switch_mode=None,
    uplink_teaming_policy_names=None,
    tags=None,
    is_default=None,
):
    """
    Creates transport zone to NSX-T Manager

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_zone.create_transport_zone hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    transport_type
        (Optional) Transport type to be added to NSX-T Manager

    description
        (Optional) Description for the transport zones

    display_name
        (Optional) Value of the transport zone which we want to specify

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    host_switch_id
       (Optional) The host switch id generated by the system.

    host_switch_mode
       (Optional) Operational mode of the transport zone.
       STANDARD mode applies to all the hypervisors. ENS mode stands for Enhanced Networking Stack.
       This feature is only available for ESX hypervisor. It is not available on KVM, EDGE and Public Cloud Gateway etc.
       When a Transport Zone mode is set to ENS, only Transport Nodes of type ESX can participate in such a Transport Zone.

    host_switch_name
       (Optional) Name of the host switch on all transport nodes in this transport zone that will be used to run NSX network traffic.
       If this name is unset or empty then the default host switch name will be used.

    uplink_teaming_policy_names
       (Optional) Names of the switching uplink teaming policies that are supported by this transport zone.

    tags
       (Optional) Opaque identifiers meaningful to the API user
    """
    log.info("Creating Transport Zones")
    url = TRANSPORT_ZONE_BASE_URL.format(hostname)
    req_data = common._filter_kwargs(
        allowed_kwargs=create_params_for_transport_zones,
        default_dict={},
        host_switch_name=host_switch_name,
        transport_type=transport_type,
        description=description,
        host_switch_id=host_switch_id,
        host_switch_mode=host_switch_mode,
        uplink_teaming_policy_names=uplink_teaming_policy_names,
        display_name=display_name,
        tags=tags,
        is_default=is_default,
    )
    return nsxt_request.call_api(
        method="post",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=req_data,
    )


def update(
    hostname,
    username,
    password,
    transport_zone_id,
    revision,
    verify_ssl=None,
    cert=None,
    cert_common_name=None,
    host_switch_name=None,
    transport_type=None,
    description=None,
    host_switch_id=None,
    host_switch_mode=None,
    uplink_teaming_policy_names=None,
    tags=None,
    is_default=None,
    display_name=None,
):
    """
    Update transport zone to NSX-T Manager

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_zone.update hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    transport_zone_id
        Id of the transport zone for which update needs to be done

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    revision
        Required to keep track of the total posted version

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    host_switch_id
        (Optional) The host switch id generated by the system.

    host_switch_mode
        (Optional) Operational mode of the transport zone.
        STANDARD mode applies to all the hypervisors. ENS mode stands for Enhanced Networking Stack.
        This feature is only available for ESX hypervisor. It is not available on KVM, EDGE and Public Cloud Gateway etc.
        When a Transport Zone mode is set to ENS, only Transport Nodes of type ESX can participate in such a Transport Zone.

    host_switch_name
        (Optional) Name of the host switch on all transport nodes in this transport zone that will be used to run NSX network traffic.
        If this name is unset or empty then the default host switch name will be used.

    uplink_teaming_policy_names
        (Optional) Names of the switching uplink teaming policies that are supported by this transport zone.

    tags
        (Optional) Opaque identifiers meaningful to the API user
    """
    log.info("Updating transport zones")
    url = TRANSPORT_ZONE_BASE_URL.format(hostname) + "/{}".format(transport_zone_id)
    update_body = common._filter_kwargs(
        allowed_kwargs=create_params_for_transport_zones,
        default_dict={},
        host_switch_name=host_switch_name,
        transport_type=transport_type,
        description=description,
        host_switch_id=host_switch_id,
        host_switch_mode=host_switch_mode,
        uplink_teaming_policy_names=uplink_teaming_policy_names,
        display_name=display_name,
        tags=tags,
        is_default=is_default,
    )
    update_body["_revision"] = revision
    return nsxt_request.call_api(
        method="put",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=update_body,
    )


def delete(
    hostname,
    username,
    password,
    transport_zone_id,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
):
    """
    Deletes transport-zone in NSX-T Manager

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_zone.delete_license hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    transport-zone-id
        Transport zone id which needs to be deleted

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    For more information, see `VMware API docs for NSX-T <https://code.vmware.com/apis/1163/nsx-t>`_

    """
    log.info("Deleting transport zones for %s", transport_zone_id)
    url = TRANSPORT_ZONE_BASE_URL.format(hostname) + "/{}".format(transport_zone_id)
    delete_response = nsxt_request.call_api(
        method="delete",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
    )
    return delete_response or {"message": "Deleted transport zone successfully"}



import hashlib
import logging
import socket
import ssl

import salt.exceptions
import saltext.vmware.utils.cluster as utils_cluster
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.datacenter as utils_datacenter

# pylint: disable=no-name-in-module
try:
    from pyVmomi import vim, vmodl

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

log = logging.getLogger(__name__)


def get_hosts(
    service_instance,
    datacenter_name=None,
    host_names=None,
    cluster_name=None,
    get_all_hosts=False,
):
    """
    Returns a list of vim.HostSystem objects representing ESXi hosts
    in a vcenter filtered by their names and/or datacenter, cluster membership.

    service_instance
        The Service Instance Object from which to obtain the hosts.

    datacenter_name
        The datacenter name. Default is None.

    host_names
        The host_names to be retrieved. Default is None.

    cluster_name
        The cluster name - used to restrict the hosts retrieved. Only used if
        the datacenter is set.  This argument is optional.

    get_all_hosts
        Specifies whether to retrieve all hosts in the container.
        Default value is False.
    """
    properties = ["name"]
    if cluster_name and not datacenter_name:
        raise salt.exceptions.ArgumentValueError(
            "Must specify the datacenter when specifying the cluster"
        )
    if not host_names:
        host_names = []
    if not datacenter_name:
        # Assume the root folder is the starting point
        start_point = utils_common.get_root_folder(service_instance)
    else:
        start_point = utils_datacenter.get_datacenter(service_instance, datacenter_name)
        if cluster_name:
            # Retrieval to test if cluster exists. Cluster existence only makes
            # sense if the datacenter has been specified
            properties.append("parent")

    # Search for the objects
    hosts = utils_common.get_mors_with_properties(
        service_instance,
        vim.HostSystem,
        container_ref=start_point,
        property_list=properties,
    )
    log.trace("Retrieved hosts: %s", [h["name"] for h in hosts])
    filtered_hosts = []
    for h in hosts:
        # Complex conditions checking if a host should be added to the
        # filtered list (either due to its name and/or cluster membership)

        if cluster_name:
            if not isinstance(h["parent"], vim.ClusterComputeResource):
                continue
            parent_name = utils_common.get_managed_object_name(h["parent"])
            if parent_name != cluster_name:
                continue

        if get_all_hosts:
            filtered_hosts.append(h["object"])
            continue

        if h["name"] in host_names:
            filtered_hosts.append(h["object"])
    return filtered_hosts


# TODO Support host caches on multiple datastores
def configure_host_cache(host_ref, datastore_ref, swap_size_MiB, host_cache_manager=None):
    """
    Configures the host cahe of the specified host

    host_ref
        The vim.HostSystem object representing the host that contains the
        requested disks.

    datastore_ref
        The vim.Datastore opject representing the datastore the host cache will
        be configured on.

    swap_size_MiB
        The size in Mibibytes of the swap.

    host_cache_manager
        The vim.HostCacheConfigurationManager object representing the cache
        configuration manager on the specified host. Default is None. If None,
        it will be retrieved in the method
    """
    hostname = utils_common.get_managed_object_name(host_ref)
    if not host_cache_manager:
        props = utils_common.get_properties_of_managed_object(
            host_ref, ["configManager.cacheConfigurationManager"]
        )
        if not props.get("configManager.cacheConfigurationManager"):
            raise salt.exceptions.VMwareObjectRetrievalError(
                "Host '{}' has no host cache".format(hostname)
            )
        host_cache_manager = props["configManager.cacheConfigurationManager"]
    log.trace(
        "Configuring the host cache on host '%s', datastore '%s', " "swap size=%s MiB",
        hostname,
        datastore_ref.name,
        swap_size_MiB,
    )

    spec = vim.HostCacheConfigurationSpec(datastore=datastore_ref, swapSize=swap_size_MiB)
    log.trace("host_cache_spec=%s", spec)
    try:
        task = host_cache_manager.ConfigureHostCache_Task(spec)
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    utils_common.wait_for_task(task, hostname, "HostCacheConfigurationTask")
    log.trace("Configured host cache on host '%s'", hostname)
    return True


def list_hosts(service_instance):
    """
    Returns a list of hosts associated with a given service instance.

    service_instance
        The Service Instance Object from which to obtain hosts.
    """
    return utils_common.list_objects(service_instance, vim.HostSystem)


def disconnect_host(host, service_instance):
    """
    Disconnects host from vCenter instance

    Returns connection state of host

    host
        Name of ESXi instance in vCenter.

    service_instance
        The Service Instance Object from which to obtain host.
    """
    host = utils_common.get_mor_by_property(service_instance, vim.HostSystem, host)
    if host.summary.runtime.connectionState == "disconnected":
        return host.summary.runtime.connectionState
    task = host.DisconnectHost_Task()
    host = utils_common.wait_for_task(task, host, "disconnect host task")
    return host.summary.runtime.connectionState


def reconnect_host(host, service_instance):
    """
    Reconnects host from vCenter instance

    Returns connection state of host

    host
        Name of ESXi instance in vCenter.

    service_instance
        The Service Instance Object from which to obtain host.
    """
    host = utils_common.get_mor_by_property(service_instance, vim.HostSystem, host)
    if host.summary.runtime.connectionState == "connected":
        return host.summary.runtime.connectionState
    task = host.ReconnectHost_Task()
    ret_host = utils_common.wait_for_task(task, host, "reconnect host task")
    return ret_host.summary.runtime.connectionState


def move_host(host, cluster_name, service_instance):
    """
    Move host to a different cluster.

    Returns connection state of host

    host
        Name of ESXi instance in vCenter.

    cluster_name
        Name of cluster to move host to.

    service_instance
        The Service Instance Object from which to obtain host.
    """
    host_ref = utils_common.get_mor_by_property(service_instance, vim.HostSystem, host)
    cluster_ref = utils_common.get_mor_by_property(
        service_instance, vim.ClusterComputeResource, cluster_name
    )
    host_dc = utils_common.get_parent_of_type(host_ref, vim.Datacenter)
    host_cluster = utils_common.get_parent_of_type(host_ref, vim.ClusterComputeResource)
    cluster_dc = utils_common.get_parent_of_type(cluster_ref, vim.Datacenter)
    if host_dc != cluster_dc:
        raise salt.exceptions.VMwareApiError("Cluster has to be in the same datacenter")
    task = cluster_ref.MoveInto_Task([host_ref])
    utils_common.wait_for_task(task, cluster_name, "move host task")
    return f"moved {host} from {host_cluster.name} to {cluster_ref.name}"


def remove_host(host, service_instance):
    """
    Removes host from vCenter instance.

    Returns connection state of host

    host
        Name of ESXi instance in vCenter.

    service_instance
        The Service Instance Object from which to obtain host.
    """
    host_ref = utils_common.get_mor_by_property(service_instance, vim.HostSystem, host)
    task = host_ref.Destroy_Task()
    utils_common.wait_for_task(task, host, "destroy host task")
    return f"removed host {host}"


def _format_ssl_thumbprint(number):
    """
    Formats ssl cert number

    number
        Number to be formatted into ssl thumbprint
    """
    string = str(number)
    return ":".join(a + b for a, b in zip(string[::2], string[1::2]))


def _get_host_thumbprint(ip, verify_host_cert=True):
    """
    Returns host's ssl thumbprint.

    ip
        IP address of host.
    """
    ctx = ssl.SSLContext()
    if verify_host_cert:
        ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
    with socket.create_connection((ip, 443)) as _socket:
        _socket.settimeout(1)
        with ctx.wrap_socket(_socket, server_hostname=ip) as wrappedSocket:

            cert = wrappedSocket.getpeercert(True)
            sha1 = hashlib.sha1(cert).hexdigest()
            response = _format_ssl_thumbprint(sha1)
            return response


def add_host(
    host,
    root_user,
    password,
    cluster_name,
    datacenter_name,
    verify_host_cert,
    connect,
    service_instance,
):
    """
    Adds host from vCenter instance

    Returns connection state of host

    host
        IP address or hostname of ESXI instance.

    root_user
        Username with root privilege to ESXi instance.

    password
        Password to root user.

    cluster_name
        Name of cluster ESXi host is being added to.

    datacenter
        Datacenter that contains cluster that ESXi instance is being added to.

    verify_host_cert
        Validates the host's SSL certificate is signed by a CA, and that the hostname in the certificate matches the host.

    connect
        Specifies whether host should be connected after being added.

    service_instance
        The Service Instance Object to place host on.
    """
    dc_ref = utils_common.get_datacenter(service_instance, datacenter_name)
    cluster_ref = utils_cluster.get_cluster(dc_ref, cluster_name)

    connect_spec = vim.host.ConnectSpec()
    connect_spec.sslThumbprint = _get_host_thumbprint(host, verify_host_cert)
    connect_spec.hostName = host
    connect_spec.userName = root_user
    connect_spec.password = password

    task = cluster_ref.AddHost_Task(connect_spec, connect)
    host_ref = utils_common.wait_for_task(task, host, "add host task")
    return host_ref.summary.runtime.connectionState


def get_host(host, service_instance):
    return utils_common.get_mor_by_property(service_instance, vim.HostSystem, host)


def get_firewall_config(
    ruleset_name,
    host_name,
    service_instance,
):
    """
    Get Firewall a rule configuration on matching ESXi hosts.

    ruleset_name
        Name of firewall rule.

    host_name
        Filter by this ESXi hostname (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    """
    host = get_host(
        host=host_name,
        service_instance=service_instance,
    )

    firewall_config = host.configManager.firewallSystem
    if not firewall_config:
        return None
    for ruleset in firewall_config.firewallInfo.ruleset:
        if ruleset_name == ruleset.key:
            ret = {
                host.name: {
                    ruleset.key: {
                        "allowed_hosts": {
                            "ip_address": list(ruleset.allowedHosts.ipAddress),
                            "all_ip": ruleset.allowedHosts.allIp,
                            "ip_network": [
                                "{}/{}".format(ip.network, ip.prefixLength)
                                for ip in ruleset.allowedHosts.ipNetwork
                            ],
                        },
                        "service": ruleset.service,
                        "enabled": ruleset.enabled,
                        "rule": [
                            {
                                "port": r.port,
                                "end_port": r.endPort,
                                "direction": r.direction,
                                "port_type": r.portType,
                                "protocol": r.protocol,
                            }
                            for r in ruleset.rule
                        ],
                    }
                }
            }
    return ret


def get_host_ssds(host_reference):
    """
    Helper function that returns a list of ssd objects for a given host.
    """
    return get_host_disks(host_reference).get("SSDs")


def get_host_non_ssds(host_reference):
    """
    Helper function that returns a list of Non-SSD objects for a given host.
    """
    return get_host_disks(host_reference).get("Non-SSDs")


def get_host_disks(host_reference):
    """
    Helper function that returns a dictionary containing a list of SSD and Non-SSD disks.
    """
    storage_system = host_reference.configManager.storageSystem
    disks = storage_system.storageDeviceInfo.scsiLun
    ssds = []
    non_ssds = []

    for disk in disks:
        try:
            has_ssd_attr = disk.ssd
        except AttributeError:
            has_ssd_attr = False
        if has_ssd_attr:
            ssds.append(disk)
        else:
            non_ssds.append(disk)

    return {"SSDs": ssds, "Non-SSDs": non_ssds}


def get_missing_firewall_rules(rules, hosts):
    missing = []
    for host in hosts:
        ruleset = host.configManager.firewallSystem.firewallInfo.ruleset
        existing = [r.key for r in ruleset]
        for rule in rules:
            if rule["name"] not in existing:
                missing.append((rule.key, host.name))
    return missing



"""
    NSX-T API URLs
"""

_SITE_URL = "/infra/sites"
_DOMAIN_URL = "/infra/domains"
_ENFORCEMENT_POINT_URL = _SITE_URL + "/{}/enforcement-points"

TIER_0_URL = "/infra/tier-0s"
TIER_0_STATIC_ROUTE_URL = TIER_0_URL + "/{}/static-routes"
TIER_0_LOCALE_SERVICE_URL = TIER_0_URL + "/{}/locale-services"
TIER_0_LS_INTERFACE_URL = TIER_0_LOCALE_SERVICE_URL + "/{}/interfaces"
TIER_0_BGP_NEIGHBOR_URL = TIER_0_LOCALE_SERVICE_URL + "/{}/bgp/neighbors"
TIER_0_BFD_PEERS = TIER_0_STATIC_ROUTE_URL + "/bfd-peers"

IPV6_DAD_PROFILE_URL = "/infra/ipv6-dad-profiles"
IPV6_NDRA_PROFILE_URL = "/infra/ipv6-ndra-profiles"
DHCP_RELAY_CONFIG_URL = "/infra/dhcp-relay-configs"
BFD_PROFILE_URL = "/infra/bfd-profiles"
EDGE_CLUSTER_URL = _ENFORCEMENT_POINT_URL + "/{}/edge-clusters"
EDGE_NODE_URL = EDGE_CLUSTER_URL + "/{}/edge-nodes"
SEGMENT_URL = "/infra/segments"
SEGMENT_PORT_URL = SEGMENT_URL + "/{}/ports"
TRANSPORT_ZONE_URL = _ENFORCEMENT_POINT_URL + "/{}/transport-zones"
IP_POOL_URL = "/infra/ip-pools"
TRANSPORT_ZONE_URL = _ENFORCEMENT_POINT_URL + "/{}/transport-zones"
IP_POOL_URL = "/infra/ip-pools"

TIER_1_URL = "/infra/tier-1s"
TIER_1_STATIC_ROUTE_URL = TIER_1_URL + "/{}/static-routes"
TIER_1_LOCALE_SERVICE_URL = TIER_1_URL + "/{}/locale-services"
TIER_1_LS_INTERFACE_URL = TIER_1_LOCALE_SERVICE_URL + "/{}/interfaces"
TIER_1_BGP_NEIGHBOR_URL = TIER_1_LOCALE_SERVICE_URL + "/{}/bgp/neighbors"



"""
Take changes data from salt and merge different sections in single one

Add the following to the minion or master configuration file.

.. code-block:: yaml

    merge_changes.filename: <path_to_output_file>

Default is ``/var/log/salt/changes``.
"""
import logging

import salt.returners
import salt.utils.files
import salt.utils.json

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "merge_changes_esxi"


def __virtual__():
    return __virtualname__


def _get_options(ret):
    """
    Returns options used for the merge_changes returner.
    """
    defaults = {"filename": "/var/log/salt/changes"}
    attrs = {"filename": "filename"}
    _options = salt.returners.get_returner_options(
        __virtualname__,
        ret,
        attrs,
        __salt__=__salt__,
        __opts__=__opts__,
        defaults=defaults,
    )

    return _options


def returner(ret):
    """
    Write the return data to a file on the minion.
    """
    opts = _get_options(ret)
    return_value = ret["return"]
    changes = {}
    for key in return_value:
        state = return_value[key]
        if key.startswith("vmware_esxi_"):
            if "esxi" not in changes:
                changes["esxi"] = {}
            for host in state["changes"]:
                if host not in changes["esxi"]:
                    changes["esxi"][host] = {}
                changes["esxi"][host][state["__id__"]] = state["changes"][host]["old"]
        else:
            changes[state["__id__"]] = state["changes"]
    with salt.utils.files.flopen(opts["filename"], "a") as logfile:
        salt.utils.json.dump(changes, logfile)
        logfile.write("\n")



"""
NSX-T Datacenter License Management state module
================================================

:maintainer: <VMware>
:maturity: new

Add new license and delete existing license from NSX-T Manager.

Example usage :

.. code-block:: yaml

    Ensure license exists:
        nsxt_license.present:
            hostname: 203.0.113.13
            username: admin
            password: admin_password
            license_key: ABCDE-12345-ABCDE-12345-ABCDE
            cert: /path/to/client/certificate

.. warning::

    It is recommended to pass the NSX authentication details using Pillars rather than specifying as plain text in SLS
    files.


"""


def __virtual__():
    """
    Only load if the nsxt_license module is available in __salt__
    """
    return (
        "nsxt_license" if "nsxt_license.get_licenses" in __salt__ else False,
        "'nsxt_license' binary not found on system",
    )


def present(
    name,
    hostname,
    username,
    password,
    license_key,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
):
    """
    Ensure a given license exists on given NSX-T Manager

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    license_key
        The license key to be added to NSX-T Manager

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    get_licenses_result = __salt__["nsxt_license.get_licenses"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    if "error" in get_licenses_result:
        ret["result"] = False
        ret["comment"] = "Failed to get current licenses from NSX-T Manager : {}".format(
            get_licenses_result["error"]
        )
        return ret

    if get_licenses_result["result_count"] != 0:
        results = get_licenses_result["results"]
        for result in results:
            if result["license_key"] == license_key:
                ret["result"] = True
                ret["comment"] = "License key is already present"
                return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "License would be added to NSX-T Manager"
        return ret

    apply_license_result = __salt__["nsxt_license.apply_license"](
        hostname=hostname,
        username=username,
        password=password,
        license_key=license_key,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    if "error" in apply_license_result:
        ret["result"] = False
        ret["comment"] = "Failed to apply license to NSX-T Manager : {}".format(
            apply_license_result["error"]
        )
        return ret

    get_licenses_result_after_apply = __salt__["nsxt_license.get_licenses"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    if "error" in get_licenses_result_after_apply:
        ret["result"] = False
        ret[
            "comment"
        ] = "Failed to retrieve licenses after applying current license from NSX-T Manager : {}".format(
            get_licenses_result_after_apply["error"]
        )
        return ret

    ret["comment"] = "License added successfully"
    ret["changes"]["old"] = get_licenses_result
    ret["changes"]["new"] = get_licenses_result_after_apply
    ret["result"] = True

    return ret


def absent(
    name,
    hostname,
    username,
    password,
    license_key,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
):
    """
    Ensure a given license does not exist on given NSX-T Manager

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    license_key
        The license key to be removed from NSX-T Manager

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    get_licenses_result = __salt__["nsxt_license.get_licenses"](
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    if "error" in get_licenses_result:
        ret["result"] = False
        ret["comment"] = "Failed to get current licenses from NSX-T Manager : {}".format(
            get_licenses_result["error"]
        )
        return ret

    if get_licenses_result["result_count"] != 0:
        results = get_licenses_result["results"]
        for result in results:
            if result["license_key"] == license_key:
                if __opts__["test"]:
                    ret["result"] = None
                    ret["comment"] = "License would be removed from NSX-T Manager"
                    return ret

                delete_license_result = __salt__["nsxt_license.delete_license"](
                    hostname=hostname,
                    username=username,
                    password=password,
                    license_key=license_key,
                    verify_ssl=verify_ssl,
                    cert=cert,
                    cert_common_name=cert_common_name,
                )

                if "error" in delete_license_result:
                    ret["result"] = False
                    ret["comment"] = "Failed to delete license from NSX-T Manager : {}".format(
                        delete_license_result["error"]
                    )
                    return ret

                get_licenses_result_after_delete = __salt__["nsxt_license.get_licenses"](
                    hostname=hostname,
                    username=username,
                    password=password,
                    verify_ssl=verify_ssl,
                    cert=cert,
                    cert_common_name=cert_common_name,
                )

                if "error" in get_licenses_result_after_delete:
                    ret["result"] = False
                    ret["comment"] = (
                        "Failed to retrieve licenses "
                        "after deleting current license from NSX-T Manager : {}".format(
                            get_licenses_result_after_delete["error"]
                        )
                    )
                    return ret

                ret["comment"] = "License removed successfully"
                ret["changes"]["old"] = get_licenses_result
                ret["changes"]["new"] = get_licenses_result_after_delete
                ret["result"] = True
                return ret

    ret["result"] = True
    ret["comment"] = "License key is not present in NSX-T Manager"
    return ret



"""
NSX-T Transport_Zone state module
"""
import logging

log = logging.getLogger(__name__)

__virtualname__ = "nsxt_transport_zone"

try:
    from saltext.vmware.modules import nsxt_transport_zone

    HAS_NSXT_TRANSPORT_ZONE = True
except ImportError:
    HAS_NSXT_TRANSPORT_ZONE = False


def __virtual__():
    if not HAS_NSXT_TRANSPORT_ZONE:
        return False, "'nsxt_transport_zone' binary not found on system"
    return "nsxt_transport_zone"


def _needs_update(transport_zone_dict, **transport_zone_param):
    updatable_params = {
        "is_default",
        "description",
        "uplink_teaming_policy_names",
        "host_switch_id",
        "host_switch_mode",
        "host_switch_name",
        "transport_type",
        "tags",
    }

    for param in updatable_params:
        param_val_in_existing_transport_zone = transport_zone_dict.get(param)
        param_val_in_inputs_for_update = transport_zone_param.get(param)

        if not param_val_in_existing_transport_zone and param_val_in_inputs_for_update:
            return True
        if param_val_in_existing_transport_zone and not param_val_in_inputs_for_update:
            return True
        if (
            param_val_in_existing_transport_zone
            and param_val_in_inputs_for_update
            and param_val_in_existing_transport_zone != param_val_in_inputs_for_update
        ):
            return True


def present(
    name,
    hostname,
    username,
    password,
    display_name,
    host_switch_name,
    transport_type,
    is_default=None,
    description=None,
    verify_ssl=None,
    cert=None,
    cert_common_name=None,
    host_switch_id=None,
    host_switch_mode=None,
    uplink_teaming_policy_names=None,
    tags=None,
):
    """

    Registers transport zone in NSX-T Manager or updates the transport zone

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_zone.present hostname=nsxt-manager.local username=admin ...

    name
        name of the operation to perform

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    host-switch-name
        Host switch name for the transport zone

    transport-type
        Transport type for the transport zone

    is_default
        (Optional) Flag to indicate if the transport zone is the default one.
        Only one transport zone can be the default one for a given transport zone type.
        APIs that need transport zone can choose to use the default transport zone if a transport zone is not given

    display_name
        Display name for the transport zone

    description
        (Optional) Description for the transport zone

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    host_switch_id
        (Optional) The host switch id generated by the system.

    host_switch_mode
        (Optional) Operational mode of the transport zone.
        STANDARD mode applies to all the hypervisors. ENS mode stands for Enhanced Networking Stack.
        This feature is only available for ESX hypervisor. It is not available on KVM, EDGE and Public Cloud Gateway etc.
        When a Transport Zone mode is set to ENS, only Transport Nodes of type ESX can participate in such a Transport Zone.

    host_switch_name
        (Optional) Name of the host switch on all transport nodes in this transport zone that will be used to run NSX network traffic.
        If this name is unset or empty then the default host switch name will be used.

    uplink_teaming_policy_names
        (Optional) Names of the switching uplink teaming policies that are supported by this transport zone.

    tags
        (Optional) Opaque identifiers meaningful to the API user

    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    get_transport_zone = __salt__["nsxt_transport_zone.get_by_display_name"](
        hostname=hostname,
        username=username,
        password=password,
        display_name=display_name,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    transport_zone_dict, transport_zone_id = None, None
    is_update = False

    if "error" in get_transport_zone:
        ret["result"] = False
        ret["comment"] = "Failed to get the transport zones : {}".format(
            get_transport_zone["error"]
        )
        return ret

    transport_zone_response_by_display_name = get_transport_zone["results"]

    transport_zone_dict = (
        transport_zone_response_by_display_name[0]
        if len(transport_zone_response_by_display_name) > 0
        else None
    )

    if len(transport_zone_response_by_display_name) > 1:
        ret["result"] = False
        ret["comment"] = "More than one transport zone exist with same display name : {}".format(
            display_name
        )
        return ret

    if transport_zone_dict is not None:
        is_update = _needs_update(
            transport_zone_dict,
            host_switch_name=host_switch_name,
            transport_type=transport_type,
            description=description,
            host_switch_id=host_switch_id,
            host_switch_mode=host_switch_mode,
            uplink_teaming_policy_names=uplink_teaming_policy_names,
            tags=tags,
            is_default=is_default,
        )
        transport_zone_id = transport_zone_dict["id"]
        revision = transport_zone_dict["_revision"]

    if __opts__["test"]:
        if transport_zone_dict is None:
            ret["result"] = None
            ret["comment"] = "Transport zone will be created in NSX-T Manager"
        else:
            ret["result"] = None
            ret["comment"] = "Transport zone would be updated"
        return ret

    if not is_update:

        if transport_zone_id:
            ret["result"] = True
            ret["comment"] = "Transport zone with display_name %s already exists", display_name
            return ret

        else:
            log.info("Start of the create of the transport zone")
            create_transport_zone = __salt__["nsxt_transport_zone.create"](
                hostname=hostname,
                username=username,
                password=password,
                verify_ssl=verify_ssl,
                cert=cert,
                cert_common_name=cert_common_name,
                display_name=display_name,
                host_switch_name=host_switch_name,
                transport_type=transport_type,
                description=description,
                host_switch_id=host_switch_id,
                host_switch_mode=host_switch_mode,
                uplink_teaming_policy_names=uplink_teaming_policy_names,
                tags=tags,
                is_default=is_default,
            )

        if "error" in create_transport_zone:
            ret["result"] = False
            ret["comment"] = "Fail to create transport_zone : {}".format(
                create_transport_zone["error"]
            )
            return ret

        ret["comment"] = "Transport Zone created successfully"
        ret["result"] = True
        ret["changes"]["new"] = create_transport_zone
        return ret
    else:
        log.info("Start of the update of the transport zone")
        update_transport_zone = __salt__["nsxt_transport_zone.update"](
            hostname=hostname,
            username=username,
            password=password,
            verify_ssl=verify_ssl,
            cert=cert,
            cert_common_name=cert_common_name,
            transport_zone_id=transport_zone_id,
            revision=revision,
            host_switch_name=host_switch_name,
            transport_type=transport_type,
            description=description,
            host_switch_id=host_switch_id,
            host_switch_mode=host_switch_mode,
            uplink_teaming_policy_names=uplink_teaming_policy_names,
            tags=tags,
            is_default=is_default,
            display_name=display_name,
        )

        if "error" in update_transport_zone:
            ret["result"] = False
            ret["comment"] = "Fail to update transport_zone : {}".format(
                update_transport_zone["error"]
            )
            return ret

        ret["comment"] = "Transport Zone updated successfully"
        ret["result"] = True
        ret["changes"]["old"] = transport_zone_dict
        ret["changes"]["new"] = update_transport_zone
        return ret


def absent(
    name,
    hostname,
    username,
    password,
    display_name,
    verify_ssl=None,
    cert=None,
    cert_common_name=None,
):
    """

    Deletes transport zone in NSX-T Manager if present. Requires display_name of the transport zone which user want to delete and also the tranpsort type of the transport zone

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_transport_zone.absent hostname=nsxt-manager.local username=admin ...

    name
        Name of the operation to perform

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    display_name
        display_name of the transport zone to be deleted

    transport_type
        transport_type of the transport zone to be deleted

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    """

    log.info("Start of the delete of the transport zone")

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    transport_zone_dict = None

    get_transport_zone = __salt__["nsxt_transport_zone.get_by_display_name"](
        hostname=hostname,
        username=username,
        password=password,
        display_name=display_name,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    if "error" in get_transport_zone:
        ret["result"] = False
        ret["comment"] = "Failed to get the transport zones : {}".format(
            get_transport_zone["error"]
        )
        return ret

    transport_zone_response_by_display_name = get_transport_zone["results"]

    transport_zone_dict = (
        transport_zone_response_by_display_name[0]
        if len(transport_zone_response_by_display_name) > 0
        else None
    )

    if len(transport_zone_response_by_display_name) > 1:
        ret["result"] = False
        ret["comment"] = "More than one transport zone exist with same display name : {}".format(
            display_name
        )
        return ret

    if len(transport_zone_response_by_display_name) == 0:
        ret["result"] = True
        ret["comment"] = "No transport zone exist with same display name : {}".format(display_name)
        return ret

    if transport_zone_dict is not None:
        id = transport_zone_dict["id"]

    delete_transport_zone = __salt__["nsxt_transport_zone.delete"](
        hostname=hostname,
        username=username,
        password=password,
        transport_zone_id=id,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    if "error" in delete_transport_zone:
        ret["result"] = False
        ret["comment"] = "Failed to delete the transport-zone : {}".format(
            delete_transport_zone["error"]
        )
        return ret

    ret["comment"] = "Transport zone deleted successfully"
    ret["changes"]["old"] = transport_zone_dict
    ret["changes"]["new"] = {}
    ret["result"] = True

    return ret



"""
VMC Security Rule state module

Add new security rule, update existing security rule and delete existing security rule from an SDDC.

Example usage :

.. code-block:: yaml

    ensure_security_rule:
      vmc_security_rules.present:
        - hostname: sample-nsx.vmwarevmc.com
        - refresh_key: 7jPSGSZpCa8e5Ouks4UY5cZyOtynAhF
        - authorization_host: console-stg.cloud.vmware.com
        - org_id: 10e1092f-51d0-473a-80f8-137652c39fd0
        - sddc_id: b43da080-2626-f64c-88e8-7f31d9d2c306
        - domain_id: mgw
        - rule_id: vCenter_Inbound_Rule_2
        - verify_ssl: False
        - cert: /path/to/client/certificate

.. warning::

    It is recommended to pass the VMC authentication details using Pillars rather than specifying as plain text in SLS
    files.


"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_state

log = logging.getLogger(__name__)
SECURITY_RULE_NOT_FOUND_ERROR = "could not be found"

try:
    from saltext.vmware.modules import vmc_security_rules

    HAS_SECURITY_RULES = True
except ImportError:
    HAS_SECURITY_RULES = False


def __virtual__():
    if not HAS_SECURITY_RULES:
        return False, "'vmc_security_rules' binary not found on system"
    return "vmc_security_rules"


def present(
    name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    rule_id,
    verify_ssl=True,
    cert=None,
    source_groups=None,
    destination_groups=None,
    services=None,
    scope=None,
    action=None,
    tag=None,
    logged=None,
    disabled=None,
    notes=None,
    sequence_number=None,
    tags=vmc_constants.VMC_NONE,
    display_name=None,
):
    """
    Ensure a given security rule exists for given SDDC

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the security rules should be added

    domain_id
        The domain_id for which the security rule should belong to. Possible values: mgw, cgw

    rule_id
        Id of the security_rule to be added to SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    source_groups
        (Optional) List of Source group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    destination_groups
        (Optional) List of Destination group paths.
        We need paths as duplicate names may exist for groups under different domains.
        Along with paths we support IP Address of type IPv4 and IPv6.
        IP Address can be in one of the format(CIDR, IP Address, Range of IP Address).
        In order to specify all groups, use the constant "ANY". This is case insensitive.
        If "ANY" is used, it should be the ONLY element in the group array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    Note: Both source_groups and destination_groups can not be ["ANY"] when domain_id=mgw

    services
        (Optional) Names of services. In order to specify all services, use the constant "ANY".
        This is case insensitive. If "ANY" is used, it should be the ONLY element in the services array.
        Error will be thrown if ANY is used in conjunction with other values.
        If this value is not passed, then ["ANY"] will be used by default.

    scope
        (Optional) The list of policy paths where the rule is applied LR/Edge/T0/T1/LRP etc.
        Note that a given rule can be applied on multiple LRs/LRPs.

    action
        (Optional) The action to be applied to all the services.
        Possible Values for domain_id=cgw are: ALLOW, DROP, REJECT
        Possible Values for domain_id=mgw are: ALLOW

    tag
        (Optional) Tag applied on the rule. User level field which will be printed in CLI and packet logs.

    logged
        (Optional) Enable logging flag. Flag to enable packet logging. Default is disabled.

    disabled
        (Optional) Flag to disable the rule. Default is enabled.

    notes
        (Optional) Text for additional notes on changes.

    sequence_number
        (Optional) Sequence number of the Rule.
        This field is used to resolve conflicts between multiple Rules under Security or Gateway Policy for a Domain.
        If no sequence number is specified by the user, a value of 0 is assigned by default.
        If there are multiple rules with the same sequence number then their order is not deterministic.
        If a specific order of rules is desired, then one has to specify unique sequence numbers.

    tags
        (Optional) Opaque identifiers meaningful to the user.

        .. code::

            tags='[
                {
                    "tag": "<tag-key-1>"
                    "scope": "<tag-value-1>"
                },
                {
                    "tag": "<tag-key-2>"
                    "scope": "<tag-value-2>"
                }
            ]'

    display_name
        Identifier to use when displaying entity in logs or GUI. This is applicable for only update scenario.
        For create scenario, display_name would be same as rule_id.

    Example values:

        .. code::

            {
                "display_name": "vCenter Inbound Rule"
                "sequence_number": 0,
                "source_groups": [
                    "ANY"
                ],
                "services": ["/infra/services/HTTPS"],
                "logged": false,
                "disabled": false,
                "destination_groups": [
                    "/infra/domains/mgw/groups/VCENTER"
                ],
                "scope": [
                    "/infra/tier-1s/mgw"
                ],
                "action": "ALLOW",
                "tag": "",
                "notes": "",
                "tags": null
            }

    """

    input_dict = {
        "source_groups": source_groups,
        "destination_groups": destination_groups,
        "services": services,
        "scope": scope,
        "action": action,
        "tag": tag,
        "logged": logged,
        "disabled": disabled,
        "notes": notes,
        "sequence_number": sequence_number,
        "tags": tags,
        "display_name": display_name,
    }

    input_dict = {k: v for k, v in input_dict.items() if v != vmc_constants.VMC_NONE}

    get_security_rule = __salt__["vmc_security_rules.get_by_id"](
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        rule_id=rule_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    existing_security_rule = None

    if "error" not in get_security_rule:
        log.info("Security rule found with Id %s", rule_id)
        existing_security_rule = get_security_rule
    elif SECURITY_RULE_NOT_FOUND_ERROR not in get_security_rule["error"]:
        return vmc_state._create_state_response(
            name=name, comment=get_security_rule["error"], result=False
        )

    if __opts__.get("test"):
        log.info("present is called with test option")
        if existing_security_rule:
            return vmc_state._create_state_response(
                name=name, comment="State present will update Security rule {}".format(rule_id)
            )
        else:
            return vmc_state._create_state_response(
                name=name, comment="State present will create Security rule {}".format(rule_id)
            )

    if existing_security_rule:
        updatable_keys = input_dict.keys()
        is_update_required = vmc_state._check_for_updates(
            existing_security_rule, input_dict, updatable_keys, ["tags"]
        )

        if is_update_required:
            updated_security_rule = __salt__["vmc_security_rules.update"](
                hostname=hostname,
                refresh_key=refresh_key,
                authorization_host=authorization_host,
                org_id=org_id,
                sddc_id=sddc_id,
                domain_id=domain_id,
                rule_id=rule_id,
                verify_ssl=verify_ssl,
                cert=cert,
                source_groups=source_groups,
                destination_groups=destination_groups,
                services=services,
                scope=scope,
                action=action,
                tag=tag,
                logged=logged,
                disabled=disabled,
                notes=notes,
                sequence_number=sequence_number,
                tags=tags,
                display_name=display_name,
            )

            if "error" in updated_security_rule:
                return vmc_state._create_state_response(
                    name=name, comment=updated_security_rule["error"], result=False
                )

            get_security_rule_after_update = __salt__["vmc_security_rules.get_by_id"](
                hostname=hostname,
                refresh_key=refresh_key,
                authorization_host=authorization_host,
                org_id=org_id,
                sddc_id=sddc_id,
                domain_id=domain_id,
                rule_id=rule_id,
                verify_ssl=verify_ssl,
                cert=cert,
            )

            if "error" in get_security_rule_after_update:
                return vmc_state._create_state_response(
                    name=name, comment=get_security_rule_after_update["error"], result=False
                )

            return vmc_state._create_state_response(
                name=name,
                comment="Updated Security rule {}".format(rule_id),
                old_state=existing_security_rule,
                new_state=get_security_rule_after_update,
                result=True,
            )
        else:
            log.info("All fields are same as existing Security rule %s", rule_id)
            return vmc_state._create_state_response(
                name=name, comment="Security rule exists already, no action to perform", result=True
            )
    else:
        log.info("No Security rule found with Id %s", rule_id)
        created_security_rule = __salt__["vmc_security_rules.create"](
            hostname=hostname,
            refresh_key=refresh_key,
            authorization_host=authorization_host,
            org_id=org_id,
            sddc_id=sddc_id,
            domain_id=domain_id,
            rule_id=rule_id,
            verify_ssl=verify_ssl,
            cert=cert,
            source_groups=source_groups,
            destination_groups=destination_groups,
            services=services,
            scope=scope,
            action=action,
            tag=tag,
            logged=logged,
            disabled=disabled,
            notes=notes,
            sequence_number=sequence_number,
            tags=tags,
        )

        if "error" in created_security_rule:
            return vmc_state._create_state_response(
                name=name, comment=created_security_rule["error"], result=False
            )

        return vmc_state._create_state_response(
            name=name,
            comment="Created Security rule {}".format(rule_id),
            new_state=created_security_rule,
            result=True,
        )


def absent(
    name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    rule_id,
    verify_ssl=True,
    cert=None,
):
    """
    Ensure a given security rule does not exist on given SDDC

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC from which the security rule should be deleted

    domain_id
        The domain_id for which the security rule should belong to. Possible values: mgw, cgw

    rule_id
        Id of the security_rule to be deleted from SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Checking if Security rule with Id %s is present", rule_id)
    get_security_rule = __salt__["vmc_security_rules.get_by_id"](
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        rule_id=rule_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    existing_security_rule = None

    if "error" not in get_security_rule:
        log.info("Security rule found with Id %s", rule_id)
        existing_security_rule = get_security_rule
    elif SECURITY_RULE_NOT_FOUND_ERROR not in get_security_rule["error"]:
        return vmc_state._create_state_response(
            name=name, comment=get_security_rule["error"], result=False
        )

    if __opts__.get("test"):
        log.info("absent is called with test option")
        if existing_security_rule:
            return vmc_state._create_state_response(
                name=name,
                comment="State absent will delete Security rule with Id {}".format(rule_id),
            )
        else:
            return vmc_state._create_state_response(
                name=name,
                comment="State absent will do nothing as no Security rule found with Id {}".format(
                    rule_id
                ),
            )

    if existing_security_rule:
        log.info("Security rule found with Id %s", rule_id)
        deleted_security_rule = __salt__["vmc_security_rules.delete"](
            hostname=hostname,
            refresh_key=refresh_key,
            authorization_host=authorization_host,
            org_id=org_id,
            sddc_id=sddc_id,
            domain_id=domain_id,
            rule_id=rule_id,
            verify_ssl=verify_ssl,
            cert=cert,
        )

        if "error" in deleted_security_rule:
            return vmc_state._create_state_response(
                name=name, comment=deleted_security_rule["error"], result=False
            )

        return vmc_state._create_state_response(
            name=name,
            comment="Deleted Security rule {}".format(rule_id),
            old_state=existing_security_rule,
            result=True,
        )
    else:
        log.info("No Security rule found with Id %s", rule_id)
        return vmc_state._create_state_response(
            name=name, comment="No Security rule found with Id {}".format(rule_id), result=True
        )



"""
Salt execution module for VMC vCenter stats
Provides methods to get monitored items list and to query monitored item data for given vCenter.
To get the data from this module, admin credentials for vCenter are required.
"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request
from saltext.vmware.utils import vmc_vcenter_request

log = logging.getLogger(__name__)

__virtualname__ = "vmc_vcenter_stats"


def __virtual__():
    return __virtualname__


def list_monitored_items(hostname, username, password, verify_ssl=True, cert=None):
    """
    Retrieves monitored items list for given vCenter.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vcenter_stats.list_monitored_items hostname=sample-vcenter.vmwarevmc.com ...

    hostname
        Hostname of the vCenter console

    username
        admin username required to login to vCenter console

    password
       admin password required to login to vCenter console

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving the monitored items list for vCenter")
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = "{base_url}api/appliance/monitoring"
    api_url = api_url.format(base_url=api_url_base)

    headers = vmc_vcenter_request.get_headers(hostname, username, password)
    return vmc_vcenter_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        headers=headers,
        description="vmc_vcenter_stats.list_monitored_items",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def query_monitored_items(
    hostname,
    username,
    password,
    start_time,
    end_time,
    interval,
    aggregate_function,
    monitored_items,
    verify_ssl=True,
    cert=None,
):
    """
    Retrieves aggregate monitoring data for the given ``monitored_items`` across the time range.
    Data will be grouped using the ``aggregate_function`` for each ``interval`` in the time range.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_vcenter_stats.query_monitored_items hostname=sample-vcenter.vmwarevmc.com ...

    hostname
        Hostname of the vCenter console

    username
        admin username required to login to vCenter console

    password
       admin password required to login to vCenter console

    start_time
        Start time in UTC (inclusive). Ex: 2021-05-06T22:13:05.651Z

    end_time
        End time in UTC (inclusive). Ex: 2021-05-10T22:13:05.651Z

    interval
         interval between the values in hours and mins, for which aggregation will apply.
         Possible values: MINUTES30, HOURS2, MINUTES5, DAY1, HOURS6

    aggregate_function
        aggregation function. Possible values: COUNT, MAX, AVG, MIN

    monitored_items
        List of monitored item IDs. Ex: [cpu.util, mem.util]

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    msg = "Retrieving the vCenter monitoring data for {}".format(monitored_items)
    log.info(msg)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = "{base_url}api/appliance/monitoring/query"
    api_url = api_url.format(base_url=api_url_base)

    headers = vmc_vcenter_request.get_headers(hostname, username, password)
    params = {
        "start_time": start_time,
        "end_time": end_time,
        "function": aggregate_function,
        "interval": interval,
        "names": monitored_items,
    }

    return vmc_vcenter_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        headers=headers,
        description="vmc_vcenter_stats.query_monitored_items",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )



"""
Salt execution module for VMC DHCP Profiles
Provides methods to Create, Read, Update and Delete DHCP Profiles.
"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request
from saltext.vmware.utils import vmc_templates

log = logging.getLogger(__name__)

__virtualname__ = "vmc_dhcp_profiles"


def __virtual__():
    return __virtualname__


def _create_payload_for_dhcp_profile(dhcp_profile_id, profile_type, user_input):
    """
    This function creates the payload for dhcp profile based on the json template and user input passed
    """
    template_data = getattr(
        vmc_templates,
        "create_dhcp_" + profile_type + "_profiles",
        vmc_templates.create_dhcp_relay_profiles,
    )
    data = vmc_request.create_payload_for_request(template_data, user_input)
    data["id"] = data["display_name"] = dhcp_profile_id
    return data


def _update_payload_for_dhcp_profile(profile_type, existing_data, user_input):
    """
    This function creates the payload for update dhcp profile based on the json template, existing data
    and user input passed
    """
    template_data = getattr(
        vmc_templates,
        "update_dhcp_" + profile_type + "_profiles",
        vmc_templates.update_dhcp_relay_profiles,
    )
    return vmc_request.create_payload_for_request(template_data, user_input, existing_data)


def get(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    type,
    verify_ssl=True,
    cert=None,
    sort_by=None,
    sort_ascending=None,
    page_size=None,
    cursor=None,
):
    """
    Retrieves DHCP profiles of given profile type for given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_dhcp_profiles.get hostname=nsxt-manager.local type=server ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the DHCP profiles should be retrieved

    type
        The type of DHCP profiles to be retrieved. Possible values: server, relay

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order. Enabled by default.

    page_size
        (Optional) Maximum number of results to return in this page. Default page size is 1000.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    """

    log.info("Retrieving DHCP %s profiles for SDDC %s", type, sddc_id)
    profile_type = vmc_constants.DHCP_CONFIGS.format(type)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/{profile_type}"
    )
    api_url = api_url.format(
        base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, profile_type=profile_type
    )

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["cursor", "page_size", "sort_ascending", "sort_by"],
        cursor=cursor,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_dhcp_profiles.get",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_by_id(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    type,
    dhcp_profile_id,
    verify_ssl=True,
    cert=None,
):
    """
    Retrieves given DHCP profile of given profile type for given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_dhcp_profiles.get_by_id hostname=nsxt-manager.local type=server ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the DHCP profile should be retrieved

    type
        The type of DHCP profile for which the given DHCP belongs to. Possible values: server, relay

    dhcp_profile_id
        Id of the DHCP profile to be retrieved from SDDC.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Retrieving DHCP profile %s for SDDC %s", dhcp_profile_id, sddc_id)
    profile_type = vmc_constants.DHCP_CONFIGS.format(type)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/{profile_type}/{profile_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        profile_type=profile_type,
        profile_id=dhcp_profile_id,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_dhcp_profiles.get_by_id",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def delete(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    type,
    dhcp_profile_id,
    verify_ssl=True,
    cert=None,
):
    """
    Deletes given DHCP profile from the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_dhcp_profiles.delete hostname=nsxt-manager.local type=server ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC from which the DHCP profile should be deleted

    type
        The type of DHCP profile for which the given dhcp belongs to. Possible values: server, relay

    dhcp_profile_id
        Id of the DHCP profile to be deleted from SDDC.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """

    log.info("Deleting DHCP profile %s for SDDC %s", dhcp_profile_id, sddc_id)
    profile_type = vmc_constants.DHCP_CONFIGS.format(type)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/{profile_type}/{profile_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        profile_type=profile_type,
        profile_id=dhcp_profile_id,
    )

    return vmc_request.call_api(
        method=vmc_constants.DELETE_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_dhcp_profiles.delete",
        responsebody_applicable=False,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def create(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    type,
    dhcp_profile_id,
    verify_ssl=True,
    cert=None,
    server_addresses=None,
    tags=vmc_constants.VMC_NONE,
    lease_time=None,
):
    """
    Creates DHCP profile of given type for the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_dhcp_profiles.create hostname=nsxt-manager.local type=server ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the DHCP profile should be added

    type
        The type of DHCP profile for which the given DHCP belongs to. Possible values: server, relay

    dhcp_profile_id
        Id of the DHCP profile to be added to SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    server_addresses
        when type is relay, this field indicates DHCP relay addresses(DHCP server IP addresses
        for DHCP relay configuration).
        Both IPv4 and IPv6 addresses are supported.

        when type is server, this field indicates DHCP server address in CIDR format.
        Both IPv4 and IPv6 address families are supported.
        Prefix length should be less than or equal to 30 for IPv4 address family
        and less than or equal to 126 for IPv6.
        When not specified, IPv4 value is auto-assigned to 100.96.0.1/30.
        Note: This field is optional only when type is server and
        is mandatory when type is relay.

    tags
        (Optional) Opaque identifiers meaningful to the user.

        .. code::

            tags='[
                {
                    "tag": "<tag-key-1>"
                    "scope": "<tag-value-1>"
                },
                {
                    "tag": "<tag-key-2>"
                    "scope": "<tag-value-2>"
                }
            ]'

    lease_time
        (Optional) IP address lease time in seconds. Minimum value is 60.
        Maximum value is 4294967295. Default value is 86400
        Note: This field is applicable only when type is server

    Example values:

        .. code::

            dhcp-server-profiles:
            {
                "server_addresses": [
                    "203.0.113.2/23"
                ],
                "tags": [
                    {
                        "tag": "tag1",
                        "scope": "scope1"
                    }
                ],
                "lease_time": 86400
            }

            dhcp-relay-profiles:
            {
                "server_addresses": [
                    "203.0.113.1"
                ],
                "tags": [
                    {
                        "tag": "tag1",
                        "scope": "scope1"
                    }
                ]
            }
    """

    log.info("Creating DHCP profile %s for SDDC %s", dhcp_profile_id, sddc_id)

    profile_type = vmc_constants.DHCP_CONFIGS.format(type)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/{profile_type}/{profile_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        profile_type=profile_type,
        profile_id=dhcp_profile_id,
    )

    allowed_dict = {"server_addresses": server_addresses, "tags": tags, "lease_time": lease_time}

    req_data = vmc_request._filter_kwargs(
        allowed_kwargs=allowed_dict.keys(), allow_none=["tags"], **allowed_dict
    )

    payload = _create_payload_for_dhcp_profile(dhcp_profile_id, type, req_data)
    return vmc_request.call_api(
        method=vmc_constants.PUT_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_dhcp_profiles.create",
        data=payload,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def update(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    type,
    dhcp_profile_id,
    verify_ssl=True,
    cert=None,
    server_addresses=None,
    tags=vmc_constants.VMC_NONE,
    lease_time=None,
    display_name=None,
):
    """
    Updates given DHCP profile for the given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_dhcp_profiles.update hostname=nsxt-manager.local type=relay ...

    hostname
        The host name of NSX-T manager

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the VMC cloud console

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the DHCP profile belongs to

    type
        The type of DHCP profile for which the given dhcp belongs to. Possible values: server, relay

    dhcp_profile_id
        Id of the DHCP profile to be updated for given SDDC.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    server_addresses
        when type is relay, this field indicates DHCP relay addresses(DHCP server IP addresses
        for DHCP relay configuration).
        Both IPv4 and IPv6 addresses are supported.

        when type is server, this field indicates DHCP server address in CIDR format.
        Both IPv4 and IPv6 address families are supported.
        Prefix length should be less than or equal to 30 for IPv4 address family
        and less than or equal to 126 for IPv6.
        When not specified, IPv4 value is auto-assigned to 100.96.0.1/30.

        Note: This field is optional only when type is server and
        is mandatory when type is relay.

    tags
        (Optional) Opaque identifiers meaningful to the user.

        .. code::

            tags='[
                {
                    "tag": "<tag-key-1>"
                    "scope": "<tag-value-1>"
                },
                {
                    "tag": "<tag-key-2>"
                    "scope": "<tag-value-2>"
                }
            ]'

    lease_time
        (Optional) IP address lease time in seconds. Minimum value is 60.
        Maximum value is 4294967295. Default value is 86400
        Note: This field is applicable only when type is server

    display_name
        Identifier to use when displaying entity in logs or GUI.

    Example values:

        .. code::

            dhcp-server-profiles:
            {
                "display_name": "dhcp-test",
                "server_addresses": [
                    "203.0.113.2/23"
                ],
                "tags": [
                    {
                        "tag": "tag1",
                        "scope": "scope1"
                    }
                ],
                "lease_time": 86400
            }

            dhcp-relay-profiles:
            {
                "display_name": "dhcp-test",
                "server_addresses": [
                    "203.0.113.1"
                ],
                "tags": [
                    {
                        "tag": "tag1",
                        "scope": "scope1"
                    }
                ]
            }

    """

    log.info("Updating DHCP profile %s for SDDC %s", dhcp_profile_id, sddc_id)
    profile_type = vmc_constants.DHCP_CONFIGS.format(type)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/{profile_type}/{profile_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        profile_type=profile_type,
        profile_id=dhcp_profile_id,
    )

    existing_data = get_by_id(
        hostname,
        refresh_key,
        authorization_host,
        org_id,
        sddc_id,
        type,
        dhcp_profile_id,
        verify_ssl,
        cert,
    )

    # fetch the dhcp profile for the given dhcp_profile_id
    if vmc_constants.ERROR in existing_data:
        return existing_data

    allowed_dict = {
        "server_addresses": server_addresses,
        "tags": tags,
        "lease_time": lease_time,
        "display_name": display_name,
    }

    req_data = vmc_request._filter_kwargs(
        allowed_kwargs=allowed_dict.keys(), allow_none=["tags"], **allowed_dict
    )

    payload = _update_payload_for_dhcp_profile(type, existing_data, req_data)
    return vmc_request.call_api(
        method=vmc_constants.PATCH_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_dhcp_profiles.update",
        responsebody_applicable=False,
        data=payload,
        verify_ssl=verify_ssl,
        cert=cert,
    )



# SPDX-License-Identifier: Apache-2.0
import logging

import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.datastore as utils_datastore
import saltext.vmware.utils.vm as utils_vm

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "vmware_vm"
__proxyenabled__ = ["vmware_vm"]


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def set_boot_manager(
    name,
    boot_order=["cdrom", "disk", "ethernet", "floppy"],
    delay=0,
    enter_bios_setup=False,
    retry_enabled=False,
    retry_delay=10000,
    efi_secure_boot_enabled=False,
    service_instance=None,
    profile=None,
):
    """
    Manage boot option for a virtual machine

    name
        The name of the virtual machine.

    boot_order
        (List of strings) Boot order of devices. Acceptable strings: cdrom, disk, ethernet, floppy

    delay
        (integer, optional) Boot delay. When powering on or resetting, delay boot order by given milliseconds. Defaults to 0.

    enter_bios_setup
        (boolean, optional) During the next boot, force entry into the BIOS setup screen. Defaults to False.

    retry_enabled
        (boolean, optional) If the VM fails to find boot device retry. Defaults to False.

    retry_delay
        (integer, optional) If the VM fails to find boot device, automatically retry after given milliseconds. Defaults to 10000.

    efi_secure_boot_enabled
        (boolean, optional) Defaults to False.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    .. code-block:: yaml

        Set Boot Manager:
          vmware_vm.set_boot_manager:
            - name: vm01
            - boot_order:
                - cdrom
                - disk
                - ethernet
            - delay: 5000
            - enter_bios_setup: False
            - retry_enabled: True
            - retry_delay: 5000
            - efi_secure_boot_enabled: False
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    vm = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, name)
    boot_order_list = utils_vm.options_order_list(vm, boot_order)
    input_opts = {
        "bootOrder": boot_order_list,
        "bootDelay": delay,
        "enterBIOSSetup": enter_bios_setup,
        "bootRetryEnabled": retry_enabled,
        "bootRetryDelay": retry_delay,
        "efiSecureBootEnabled": efi_secure_boot_enabled,
    }
    if utils_vm.compare_boot_options(input_opts, vm.config.bootOptions):
        ret["comment"] = "already configured this way"
        return ret

    ret["changes"] = utils_vm.boot_options_dif(input_opts, vm.config.bootOptions)

    if __opts__["test"]:
        ret["comment"] = "These options are set to change."
        return ret

    result = utils_vm.change_boot_options(vm, input_opts)
    ret["comment"] = result["status"]
    return ret


def snapshot_present(
    name,
    snapshot_name,
    description="",
    include_memory=False,
    quiesce=False,
    datacenter_name=None,
    service_instance=None,
    profile=None,
):
    """
    Create virtual machine snapshot.

    name
        The name of the virtual machine.

    snapshot_name
        The name for the snapshot being created. Not unique

    description
        Description for the snapshot.

    include_memory
        (boolean, optional) If TRUE, a dump of the internal state of the virtual machine (basically a memory dump) is included in the snapshot.

    quiesce
        (boolean, optional) If TRUE and the virtual machine is powered on when the snapshot is taken, VMware Tools is used to quiesce the file system in the virtual machine.

    datacenter_name
        (optional) The name of the datacenter containing the virtual machine.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

     profile
        Profile to use (optional)


    .. code-block:: yaml

        Create Snapshot:
          vmware_vm.snapshot_present:
            - name: vm01
            - snapshot_name: backup_snapshot_1
            - description: This snapshot is a backup of vm01
            - include_memory: False
            - quiesce: True
            - datacenter_name: dc1
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    if datacenter_name:
        dc_ref = utils_common.get_mor_by_property(service_instance, vim.Datacenter, datacenter_name)
        vm_ref = utils_common.get_mor_by_property(
            service_instance, vim.VirtualMachine, name, "name", dc_ref
        )
    else:
        vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, name)

    if utils_vm.get_snapshot(vm_ref, snapshot_name, None):
        ret["comment"] = "snapshot was already created"
        return ret

    if __opts__["test"]:
        ret["changes"]["new"] = f"Snapshot {snapshot_name} will be created."
        ret["comment"] = "These options are set to change."
        return ret

    snapshot = utils_vm.create_snapshot(vm_ref, snapshot_name, description, include_memory, quiesce)

    if isinstance(snapshot, vim.vm.Snapshot):
        ret["changes"]["new"] = f"Snapshot {snapshot_name} created."
        ret["comment"] = "created"
        return ret
    else:
        ret["changes"]["result"] = False
        ret["comment"] = "failed to create"
        return ret


def snapshot_absent(
    name,
    snapshot_name,
    snapshot_id=None,
    remove_children=False,
    datacenter_name=None,
    service_instance=None,
    profile=None,
):
    """
    Create virtual machine snapshot.

    name
        The name of the virtual machine.

    snapshot_name
        The name for the snapshot being destroyed. Not unique

    snapshot_id
        (optional) ID of snapshot to be destroyed.

    remove_children
        (optional, Bool) Remove snapshots below snapshot being removed in tree.

    datacenter_name
        (optional) The name of the datacenter containing the virtual machine.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    .. code-block:: yaml

        Remove Snapshot:
          vmware_vm.snapshot_present:
            - name: vm01
            - snapshot_name: backup_snapshot_1
            - snapshot_id: 1
            - remove_children: False
            - datacenter_name: dc1
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    if datacenter_name:
        dc_ref = utils_common.get_mor_by_property(service_instance, vim.Datacenter, datacenter_name)
        vm_ref = utils_common.get_mor_by_property(
            service_instance, vim.VirtualMachine, name, "name", dc_ref
        )
    else:
        vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, name)

    snapshot = utils_vm.get_snapshot(vm_ref, snapshot_name, snapshot_id)

    if snapshot is None:
        ret["comment"] = f"Snapshot {snapshot_name} doesn't exist"
        return ret

    if __opts__["test"]:
        ret["changes"]["new"] = f"Snapshot {snapshot_name} will be destroyed."
        ret["comment"] = "These options are set to change."
        return ret
    try:
        utils_vm.destroy_snapshot(snapshot.snapshot, remove_children)
        ret["changes"]["new"] = f"Snapshot {snapshot_name} destroyed."
        ret["comment"] = "destroyed"
        return ret
    except Exception as e:
        ret["changes"]["result"] = False
        ret["comment"] = str(e)
        return ret


def relocate(name, new_host_name, datastore_name, service_instance=None, profile=None):
    """
    Relocates a virtual machine to the location specified.

    name
        The name of the virtual machine to relocate.

    new_host_name
        The name of the host you want to move the virtual machine to.

    datastore_name
        The name of the datastore you want to move the virtual machine to.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    .. code-block:: yaml

        Relocate Virtual Machine:
          vmware_vm.relocate:
            - name: vm01
            - new_host_name: host1
            - datastore_name: ds01
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, name)
    datastore_match = False
    for ds in vm_ref.datastore:
        if ds.name == datastore_name:
            datastore_match = True
    if datastore_match and vm_ref.runtime.host.name == new_host_name:
        ret["comment"] = f"{name} virtual machine is already on host {new_host_name}"
        return ret
    resources = utils_common.deployment_resources(new_host_name, service_instance)
    assert isinstance(datastore_name, str)
    datastores = utils_datastore.get_datastores(
        service_instance, datastore_name=datastore_name, datacenter_name=resources["datacenter"]
    )
    datastore_ref = datastores[0] if datastores else None
    if __opts__["test"]:
        ret["changes"]["new"] = f"{name} virtual machine will be moved to host {new_host_name}"
        ret["comment"] = "These options are set to change."
        return ret
    relo = utils_vm.relocate(
        vm_ref, resources["destination_host"], datastore_ref, resources["resource_pool"]
    )
    if relo == "success":
        ret["changes"]["new"] = f"{name} virtual machine was moved to host {new_host_name}"
        ret["comment"] = "moved"
        return ret
    ret["changes"]["result"] = False
    ret["comment"] = "failed to move"
    return ret
